# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# pylint: disable=too-many-lines

"""
The base request handler used by other modules.

This should only contain the BaseRequestHandler class.
"""

from __future__ import annotations

import contextlib
import inspect
import logging
import secrets
import sys
import traceback
import uuid
from asyncio import Future
from base64 import b64decode
from collections.abc import Awaitable, Callable, Coroutine
from contextvars import ContextVar
from datetime import date, datetime, timedelta, timezone, tzinfo
from functools import cached_property, partial, reduce
from random import Random, choice as random_choice
from types import TracebackType
from typing import Any, ClassVar, Final, cast, override
from urllib.parse import SplitResult, urlsplit, urlunsplit
from zoneinfo import ZoneInfo

import elasticapm
import html2text
import orjson as json
import regex
import tornado.web
import yaml
from accept_types import get_best_match  # type: ignore[import-untyped]
from ansi2html import Ansi2HTMLConverter
from bs4 import BeautifulSoup
from dateutil.easter import easter
from elastic_transport import ApiError, TransportError
from elasticsearch import AsyncElasticsearch
from openmoji_dist import VERSION as OPENMOJI_VERSION
from redis.asyncio import Redis
from tornado.httputil import HTTPServerRequest
from tornado.iostream import StreamClosedError
from tornado.web import (
    Finish,
    GZipContentEncoding,
    HTTPError,
    MissingArgumentError,
    OutputTransform,
)

from .. import (
    EVENT_ELASTICSEARCH,
    EVENT_REDIS,
    GH_ORG_URL,
    GH_PAGES_URL,
    GH_REPO_URL,
    NAME,
    ORJSON_OPTIONS,
    pytest_is_running,
)
from .decorators import is_authorized
from .options import ColourScheme, Options
from .static_file_handling import FILE_HASHES_DICT, fix_static_path
from .themes import THEMES
from .utils import (
    ModuleInfo,
    Permission,
    add_args_to_url,
    ansi_replace,
    apply,
    backspace_replace,
    bool_to_str,
    emoji2html,
    geoip,
    hash_bytes,
    is_prime,
    ratelimit,
    str_to_bool,
)

LOGGER: Final = logging.getLogger(__name__)

TEXT_CONTENT_TYPES: Final[set[str]] = {
    "application/javascript",
    "application/json",
    "application/vnd.asozial.dynload+json",
    "application/x-ndjson",
    "application/xml",
    "application/yaml",
}

request_ctx_var: ContextVar[HTTPServerRequest] = ContextVar("current_request")


class _RequestHandler(tornado.web.RequestHandler):
    """Base for Tornado request handlers."""

    crawler: bool = False

    @override
    async def _execute(
        self, transforms: list[OutputTransform], *args: bytes, **kwargs: bytes
    ) -> None:
        request_ctx_var.set(self.request)
        return await super()._execute(transforms, *args, **kwargs)

    # pylint: disable-next=protected-access
    _execute.__doc__ = tornado.web.RequestHandler._execute.__doc__

    @property
    def apm_client(self) -> None | elasticapm.Client:
        """Get the APM client from the settings."""
        return self.settings.get("ELASTIC_APM", {}).get("CLIENT")  # type: ignore[no-any-return]

    @property
    def apm_enabled(self) -> bool:
        """Return whether APM is enabled."""
        return bool(self.settings.get("ELASTIC_APM", {}).get("ENABLED"))

    @override
    def data_received(  # noqa: D102
        self, chunk: bytes
    ) -> None | Awaitable[None]:
        pass

    data_received.__doc__ = tornado.web.RequestHandler.data_received.__doc__

    @property
    def elasticsearch(self) -> AsyncElasticsearch:
        """
        Get the Elasticsearch client from the settings.

        This is None if Elasticsearch is not enabled.
        """
        return cast(AsyncElasticsearch, self.settings.get("ELASTICSEARCH"))

    @property
    def elasticsearch_prefix(self) -> str:
        """Get the Elasticsearch prefix from the settings."""
        return self.settings.get(  # type: ignore[no-any-return]
            "ELASTICSEARCH_PREFIX", NAME
        )

    def geoip(
        self,
        ip: None | str = None,
        database: str = geoip.__defaults__[0],  # type: ignore[index]
        *,
        allow_fallback: bool = True,
    ) -> Coroutine[None, None, None | dict[str, Any]]:
        """Get GeoIP information."""
        if not ip:
            ip = self.request.remote_ip
        if not EVENT_ELASTICSEARCH.is_set():
            return geoip(ip, database)
        return geoip(
            ip, database, self.elasticsearch, allow_fallback=allow_fallback
        )

    async def get_time(self) -> datetime:
        """Get the start time of the request in the users' timezone."""
        tz: tzinfo = timezone.utc
        try:
            geoip = await self.geoip()  # pylint: disable=redefined-outer-name
        except (ApiError, TransportError):
            LOGGER.exception("Elasticsearch request failed")
            if self.apm_client:
                self.apm_client.capture_exception()  # type: ignore[no-untyped-call]
        else:
            if geoip and "timezone" in geoip:
                tz = ZoneInfo(geoip["timezone"])
        return datetime.fromtimestamp(
            self.request._start_time, tz=tz  # pylint: disable=protected-access
        )

    def is_authorized(
        self, permission: Permission, allow_cookie_auth: bool = True
    ) -> bool | None:
        """Check whether the request is authorized."""
        return is_authorized(self, permission, allow_cookie_auth)

    @override
    def log_exception(
        self,
        typ: None | type[BaseException],
        value: None | BaseException,
        tb: None | TracebackType,
    ) -> None:
        if isinstance(value, HTTPError):
            super().log_exception(typ, value, tb)
        elif typ is StreamClosedError:
            LOGGER.debug(
                "Stream closed %s",
                self._request_summary(),
                exc_info=(typ, value, tb),  # type: ignore[arg-type]
            )
        else:
            LOGGER.error(
                "Uncaught exception %s",
                self._request_summary(),
                exc_info=(typ, value, tb),  # type: ignore[arg-type]
            )

    log_exception.__doc__ = tornado.web.RequestHandler.log_exception.__doc__

    @cached_property
    def now(self) -> datetime:
        """Get the current time."""
        # pylint: disable=method-hidden
        if pytest_is_running():
            raise AssertionError("Now accessed before it was set")
        if self.request.method in self.SUPPORTED_METHODS:
            LOGGER.error("Now accessed before it was set", stacklevel=3)
        return datetime.fromtimestamp(
            self.request._start_time,  # pylint: disable=protected-access
            tz=timezone.utc,
        )

    @override
    async def prepare(self) -> None:
        """Check authorization and call self.ratelimit()."""
        # pylint: disable=invalid-overridden-method
        self.now = await self.get_time()

        if crawler_secret := self.settings.get("CRAWLER_SECRET"):
            self.crawler = crawler_secret in self.request.headers.get(
                "User-Agent", ""
            )

        if (
            self.request.method in {"GET", "HEAD"}
            and self.redirect_to_canonical_domain()
        ):
            return

        if self.request.method != "OPTIONS" and not await self.ratelimit(True):
            await self.ratelimit()

    async def ratelimit(self, global_ratelimit: bool = False) -> bool:
        """Take b1nzy to space using Redis."""
        if (
            not self.settings.get("RATELIMITS")
            or self.request.method == "OPTIONS"
            or self.is_authorized(Permission.RATELIMITS)
            or self.crawler
        ):
            return False

        if not EVENT_REDIS.is_set():
            LOGGER.warning(
                (
                    "Ratelimits are enabled, but Redis is not available. "
                    "This can happen shortly after starting the website."
                ),
            )
            raise HTTPError(503)

        if global_ratelimit:  # TODO: add to _RequestHandler
            ratelimited, headers = await ratelimit(
                self.redis,
                self.redis_prefix,
                str(self.request.remote_ip),
                bucket=None,
                max_burst=99,  # limit = 100
                count_per_period=20,  # 20 requests per second
                period=1,
                tokens=10 if self.settings.get("UNDER_ATTACK") else 1,
            )
        else:
            method = (
                "GET" if self.request.method == "HEAD" else self.request.method
            )
            if not (limit := getattr(self, f"RATELIMIT_{method}_LIMIT", 0)):
                return False
            ratelimited, headers = await ratelimit(
                self.redis,
                self.redis_prefix,
                str(self.request.remote_ip),
                bucket=getattr(
                    self,
                    f"RATELIMIT_{method}_BUCKET",
                    self.__class__.__name__.lower(),
                ),
                max_burst=limit - 1,
                count_per_period=getattr(  # request count per period
                    self,
                    f"RATELIMIT_{method}_COUNT_PER_PERIOD",
                    30,
                ),
                period=getattr(
                    self, f"RATELIMIT_{method}_PERIOD", 60  # period in seconds
                ),
                tokens=1 if self.request.method != "HEAD" else 0,
            )

        for header, value in headers.items():
            self.set_header(header, value)

        if ratelimited:
            if self.now.date() == date(self.now.year, 4, 20):
                self.set_status(420)
                self.write_error(420)
            else:
                self.set_status(429)
                self.write_error(429)

        return ratelimited

    def redirect_to_canonical_domain(self) -> bool:
        """Redirect to the canonical domain."""
        if (
            not (domain := self.settings.get("DOMAIN"))
            or not self.request.headers.get("Host")
            or self.request.host_name == domain
            or self.request.host_name.endswith((".onion", ".i2p"))
            or regex.fullmatch(r"/[\u2800-\u28FF]+/?", self.request.path)
        ):
            return False
        port = urlsplit(f"//{self.request.headers['Host']}").port
        self.redirect(
            urlsplit(self.request.full_url())
            ._replace(netloc=f"{domain}:{port}" if port else domain)
            .geturl(),
            permanent=True,
        )
        return True

    @property
    def redis(self) -> Redis[str]:
        """
        Get the Redis client from the settings.

        This is None if Redis is not enabled.
        """
        return cast("Redis[str]", self.settings.get("REDIS"))

    @property
    def redis_prefix(self) -> str:
        """Get the Redis prefix from the settings."""
        return self.settings.get(  # type: ignore[no-any-return]
            "REDIS_PREFIX", NAME
        )


class BaseRequestHandler(_RequestHandler):
    """The base request handler used by every page and API."""

    # pylint: disable=too-many-instance-attributes, too-many-public-methods

    ELASTIC_RUM_URL: ClassVar[str] = (
        f"/@apm-rum/elastic-apm-rum.umd{'' if sys.flags.dev_mode else '.min'}.js"
        "?v=5.12.0"
    )

    COMPUTE_ETAG: ClassVar[bool] = True
    ALLOW_COMPRESSION: ClassVar[bool] = True
    MAX_BODY_SIZE: ClassVar[None | int] = None
    ALLOWED_METHODS: ClassVar[tuple[str, ...]] = ("GET",)
    POSSIBLE_CONTENT_TYPES: ClassVar[tuple[str, ...]] = ()

    module_info: ModuleInfo
    # info about page, can be overridden in module_info
    title: str = "Das Asoziale Netzwerk"
    short_title: str = "Asoziales Netzwerk"
    description: str = "Die tolle Webseite des Asozialen Netzwerks"

    used_render: bool = False

    active_origin_trials: set[str]
    content_type: None | str = None
    apm_script: None | str
    nonce: str

    def _finish(
        self, chunk: None | str | bytes | dict[str, Any] = None
    ) -> Future[None]:
        if self._finished:
            raise RuntimeError("finish() called twice")

        if chunk is not None:
            self.write(chunk)

        if (  # pylint: disable=too-many-boolean-expressions
            (content_type := self.content_type)
            and (
                content_type in TEXT_CONTENT_TYPES
                or content_type.startswith("text/")
                or content_type.endswith(("+xml", "+json"))
            )
            and self._write_buffer
            and not self._write_buffer[-1].endswith(b"\n")
        ):
            self.write(b"\n")

        return super().finish()

    @override
    def compute_etag(self) -> None | str:
        """Compute ETag with Base85 encoding."""
        if not self.COMPUTE_ETAG:
            return None
        return f'"{hash_bytes(*self._write_buffer)}"'  # noqa: B907

    @override
    def decode_argument(  # noqa: D102
        self, value: bytes, name: str | None = None
    ) -> str:
        try:
            return value.decode("UTF-8", "replace")
        except UnicodeDecodeError as exc:
            err_msg = f"Invalid unicode in {name or 'url'}: {value[:40]!r}"
            LOGGER.exception(err_msg, exc_info=exc)
            raise HTTPError(400, err_msg) from exc

    @property
    def dump(self) -> Callable[[Any], str | bytes]:
        """Get the function for dumping the output."""
        yaml_subset = self.content_type in {
            "application/json",
            "application/vnd.asozial.dynload+json",
        }

        if self.content_type == "application/yaml":
            if self.now.timetuple()[2:0:-1] == (1, 4):
                yaml_subset = True
            else:
                return lambda spam: yaml.dump(
                    spam,
                    width=self.get_int_argument("yaml_width", 80, min_=80),
                )

        if yaml_subset:
            option = ORJSON_OPTIONS
            if self.get_bool_argument("pretty", False):
                option |= json.OPT_INDENT_2
            return lambda spam: json.dumps(spam, option=option)

        return lambda spam: spam

    @override
    def finish(  # noqa: D102
        self, chunk: None | str | bytes | dict[Any, Any] = None
    ) -> Future[None]:
        as_json = self.content_type == "application/vnd.asozial.dynload+json"
        as_plain_text = self.content_type == "text/plain"
        as_markdown = self.content_type == "text/markdown"

        if (
            not isinstance(chunk, bytes | str)
            or self.content_type == "text/html"
            or not self.used_render
            or not (as_json or as_plain_text or as_markdown)
        ):
            return self._finish(chunk)

        chunk = chunk.decode("UTF-8") if isinstance(chunk, bytes) else chunk

        if as_markdown:
            return self._finish(
                f"# {self.title}\n\n"
                + html2text.html2text(chunk, self.request.full_url()).strip()
            )

        soup = BeautifulSoup(chunk, features="lxml")

        if as_plain_text:
            return self._finish(soup.get_text("\n", True))

        dictionary: dict[str, object] = {
            "url": self.fix_url(),
            "title": self.title,
            "short_title": (
                self.short_title if self.title != self.short_title else None
            ),
            "body": "".join(
                str(element)
                for element in soup.find_all(name="main")[0].contents
            ).strip(),
            "scripts": [
                {"script": script.string} | script.attrs
                for script in soup.find_all("script")
            ],
            "stylesheets": [
                stylesheet.get("href").strip()
                for stylesheet in soup.find_all("link", rel="stylesheet")
            ],
            "css": "\n".join(style.string for style in soup.find_all("style")),
        }

        return self._finish(dictionary)

    finish.__doc__ = _RequestHandler.finish.__doc__

    def finish_dict(self, **kwargs: Any) -> Future[None]:
        """Finish the request with a dictionary."""
        return self.finish(kwargs)

    def fix_url(
        self,
        url: None | str | SplitResult = None,
        new_path: None | str = None,
        **query_args: None | str | bool | float,
    ) -> str:
        """
        Fix a URL and return it.

        If the URL is from another website, link to it with the redirect page,
        otherwise just return the URL with no_3rd_party appended.
        """
        if url is None:
            url = self.request.full_url()
        if isinstance(url, str):
            url = urlsplit(url)
        if url.netloc and url.netloc.lower() != self.request.host.lower():
            if (
                not self.user_settings.ask_before_leaving
                or not self.settings.get("REDIRECT_MODULE_LOADED")
            ):
                return url.geturl()
            path = "/redirect"
            query_args["to"] = url.geturl()
            url = urlsplit(self.request.full_url())
        else:
            path = url.path if new_path is None else new_path
        path = f"/{path.strip('/')}".lower()
        if path == "/lolwut":
            path = path.upper()
        if path.startswith("/soundboard/files/") or path in FILE_HASHES_DICT:
            query_args.update(
                dict.fromkeys(self.user_settings.iter_option_names())
            )
        else:
            for (
                key,
                value,
            ) in self.user_settings.as_dict_with_str_values().items():
                query_args.setdefault(key, value)
            for key, value in self.user_settings.as_dict_with_str_values(
                include_query_argument=False,
                include_body_argument=self.request.path == "/einstellungen"
                and self.get_bool_argument("save_in_cookie", False),
            ).items():
                if value == query_args[key]:
                    query_args[key] = None

        return add_args_to_url(
            urlunsplit(
                (
                    self.request.protocol,
                    self.request.host,
                    "" if path == "/" else path,
                    url.query,
                    url.fragment,
                )
            ),
            **query_args,
        )

    @classmethod
    def get_allowed_methods(cls) -> list[str]:
        """Get allowed methods."""
        methods = {"OPTIONS", *cls.ALLOWED_METHODS}
        if "GET" in cls.ALLOWED_METHODS and cls.supports_head():
            methods.add("HEAD")
        return sorted(methods)

    def get_bool_argument(
        self,
        name: str,
        default: None | bool = None,
    ) -> bool:
        """Get an argument parsed as boolean."""
        if default is not None:
            return str_to_bool(self.get_argument(name, ""), default)
        value = str(self.get_argument(name))
        try:
            return str_to_bool(value)
        except ValueError as err:
            raise HTTPError(400, f"{value} is not a boolean") from err

    def get_display_scheme(self) -> ColourScheme:
        """Get the scheme currently displayed."""
        scheme = self.user_settings.scheme
        if scheme == "random":
            return ("light", "dark")[self.now.microsecond & 1]
        return scheme

    def get_display_theme(self) -> str:
        """Get the theme currently displayed."""
        theme = self.user_settings.theme

        if theme == "default" and self.now.month == 12:
            return "christmas"

        if theme != "random":
            return theme

        ignore_themes = ("random", "christmas")

        return random_choice(  # nosec: B311
            tuple(theme for theme in THEMES if theme not in ignore_themes)
        )

    def get_error_message(self, **kwargs: Any) -> str:
        """
        Get the error message and return it.

        If the serve_traceback setting is true (debug mode is activated),
        the traceback gets returned.
        """
        if "exc_info" in kwargs and not issubclass(
            kwargs["exc_info"][0], HTTPError
        ):
            if self.settings.get("serve_traceback") or self.is_authorized(
                Permission.TRACEBACK
            ):
                return "".join(
                    traceback.format_exception(*kwargs["exc_info"])
                ).strip()
            return "".join(
                traceback.format_exception_only(*kwargs["exc_info"][:2])
            ).strip()
        if "exc_info" in kwargs and issubclass(
            kwargs["exc_info"][0], MissingArgumentError
        ):
            return cast(str, kwargs["exc_info"][1].log_message)
        return str(self._reason)

    def get_error_page_description(self, status_code: int) -> str:
        """Get the description for the error page."""
        # pylint: disable=too-many-return-statements
        # https://developer.mozilla.org/docs/Web/HTTP/Status
        if 100 <= status_code <= 199:
            return "Hier gibt es eine total wichtige Information."
        if 200 <= status_code <= 299:
            return "Hier ist alles super! 🎶🎶"
        if 300 <= status_code <= 399:
            return "Eine Umleitung ist eingerichtet."
        if 400 <= status_code <= 499:
            if status_code == 404:
                return f"{self.request.path} wurde nicht gefunden."
            if status_code == 451:
                return "Hier wäre bestimmt geiler Scheiß."
            return "Ein Client-Fehler ist aufgetreten."
        if 500 <= status_code <= 599:
            return "Ein Server-Fehler ist aufgetreten."
        raise ValueError(
            f"{status_code} is not a valid HTTP response status code."
        )

    def get_int_argument(
        self,
        name: str,
        default: None | int = None,
        *,
        max_: None | int = None,
        min_: None | int = None,
    ) -> int:
        """Get an argument parsed as integer."""
        if default is None:
            str_value = self.get_argument(name)
            try:
                value = int(str_value, base=0)
            except ValueError as err:
                raise HTTPError(400, f"{str_value} is not an integer") from err
        elif self.get_argument(name, ""):
            try:
                value = int(self.get_argument(name), base=0)
            except ValueError:
                value = default
        else:
            value = default

        if max_ is not None:
            value = min(max_, value)
        if min_ is not None:
            value = max(min_, value)

        return value

    def get_module_infos(self) -> tuple[ModuleInfo, ...]:
        """Get the module infos."""
        return self.settings.get("MODULE_INFOS") or ()

    def get_reporting_api_endpoint(self) -> None | str:
        """Get the endpoint for the Reporting API™️."""
        if not self.settings.get("REPORTING"):
            return None
        endpoint = self.settings.get("REPORTING_ENDPOINT")

        if not endpoint or not endpoint.startswith("/"):
            return endpoint

        return f"{self.request.protocol}://{self.request.host}{endpoint}"

    @override
    def get_template_namespace(self) -> dict[str, Any]:
        """
        Add useful things to the template namespace and return it.

        They are mostly needed by most of the pages (like title,
        description and no_3rd_party).
        """
        namespace = super().get_template_namespace()
        ansi2html = partial(
            Ansi2HTMLConverter(inline=True, scheme="xterm").convert, full=False
        )
        namespace.update(self.user_settings.as_dict())
        namespace.update(
            ansi2html=partial(
                reduce, apply, (ansi2html, ansi_replace, backspace_replace)
            ),
            apm_script=(
                self.settings["ELASTIC_APM"].get("INLINE_SCRIPT")
                if self.apm_enabled
                else None
            ),
            as_html=self.content_type == "text/html",
            c=self.now.date() == date(self.now.year, 4, 1)
            or str_to_bool(self.get_cookie("c", "f") or "f", False),
            canonical_url=self.fix_url(
                self.request.full_url().upper()
                if self.request.path.upper().startswith("/LOLWUT")
                else self.request.full_url().lower()
            ).split("?")[0],
            description=self.description,
            display_theme=self.get_display_theme(),
            display_scheme=self.get_display_scheme(),
            elastic_rum_url=self.ELASTIC_RUM_URL,
            fix_static=lambda path: self.fix_url(fix_static_path(path)),
            fix_url=self.fix_url,
            emoji2html=(
                emoji2html
                if self.user_settings.openmoji == "img"
                else (
                    (lambda emoji: f'<span class="openmoji">{emoji}</span>')
                    if self.user_settings.openmoji
                    else (lambda emoji: f"<span>{emoji}</span>")
                )
            ),
            form_appendix=self.user_settings.get_form_appendix(),
            GH_ORG_URL=GH_ORG_URL,
            GH_PAGES_URL=GH_PAGES_URL,
            GH_REPO_URL=GH_REPO_URL,
            keywords="Asoziales Netzwerk, Känguru-Chroniken"
            + (
                f", {self.module_info.get_keywords_as_str(self.request.path)}"
                if self.module_info  # type: ignore[truthy-bool]
                else ""
            ),
            lang="de",  # TODO: add language support
            nonce=self.nonce,
            now=self.now,
            openmoji_version=OPENMOJI_VERSION,
            settings=self.settings,
            short_title=self.short_title,
            testing=pytest_is_running(),
            title=self.title,
        )
        namespace.update(
            {
                "🥚": timedelta()
                <= self.now.date() - easter(self.now.year)
                < timedelta(days=2),
                "🦘": is_prime(self.now.microsecond),
            }
        )
        return namespace

    def get_user_id(self) -> str:
        """Get the user id saved in the cookie or create one."""
        cookie = self.get_secure_cookie(
            "user_id",
            max_age_days=90,
            min_version=2,
        )

        user_id = cookie.decode("UTF-8") if cookie else str(uuid.uuid4())

        if not self.get_secure_cookie(  # save it in cookie or reset expiry date
            "user_id", max_age_days=30, min_version=2
        ):
            self.set_secure_cookie(
                "user_id",
                user_id,
                expires_days=90,
                path="/",
                samesite="Strict",
            )

        return user_id

    def handle_accept_header(  # pylint: disable=inconsistent-return-statements
        self, possible_content_types: tuple[str, ...], strict: bool = True
    ) -> None:
        """Handle the Accept header and set `self.content_type`."""
        if not possible_content_types:
            return
        content_type = get_best_match(
            self.request.headers.get("Accept") or "*/*",
            possible_content_types,
        )
        if content_type is None:
            if strict:
                return self.handle_not_acceptable(possible_content_types)
            content_type = possible_content_types[0]
        self.content_type = content_type
        self.set_content_type_header()

    def handle_not_acceptable(
        self, possible_content_types: tuple[str, ...]
    ) -> None:
        """Only call this if we cannot respect the Accept header."""
        self.clear_header("Content-Type")
        self.set_status(406)
        raise Finish("\n".join(possible_content_types) + "\n")

    def head(self, *args: Any, **kwargs: Any) -> None | Awaitable[None]:
        """Handle HEAD requests."""
        if self.get.__module__ == "tornado.web":
            raise HTTPError(405)
        if not self.supports_head():
            raise HTTPError(501)

        kwargs["head"] = True
        return self.get(*args, **kwargs)

    @override
    def initialize(
        self,
        *,
        module_info: ModuleInfo,
        # default is true, because then empty args dicts are
        # enough to specify that the defaults should be used
        default_title: bool = True,
        default_description: bool = True,
    ) -> None:
        """
        Get title and description from the kwargs.

        If title and description are present in the kwargs,
        then they override self.title and self.description.
        """
        self.module_info = module_info
        if not default_title:
            page_info = self.module_info.get_page_info(self.request.path)
            self.title = page_info.name
            self.short_title = page_info.short_name or self.title
        if not default_description:
            self.description = self.module_info.get_page_info(
                self.request.path
            ).description

    @override
    async def options(self, *args: Any, **kwargs: Any) -> None:
        """Handle OPTIONS requests."""
        # pylint: disable=unused-argument
        self.set_header("Allow", ", ".join(self.get_allowed_methods()))
        self.set_status(204)
        await self.finish()

    def origin_trial(self, token: bytes | str) -> bool:
        """Enable an experimental feature."""
        # pylint: disable=protected-access
        payload = json.loads(b64decode(token)[69:])
        if payload["feature"] in self.active_origin_trials:
            return True
        origin = urlsplit(payload["origin"])
        url = urlsplit(self.request.full_url())
        if url.port is None and url.scheme in {"http", "https"}:
            url = url._replace(
                netloc=f"{url.hostname}:{443 if url.scheme == 'https' else 80}"
            )
        if self.request._start_time > payload["expiry"]:
            return False
        if url.scheme != origin.scheme:
            return False
        if url.netloc != origin.netloc and not (
            payload.get("isSubdomain")
            and url.netloc.endswith(f".{origin.netloc}")
        ):
            return False
        self.add_header("Origin-Trial", token)
        self.active_origin_trials.add(payload["feature"])
        return True

    @override
    async def prepare(self) -> None:
        """Check authorization and call self.ratelimit()."""
        await super().prepare()

        if self._finished:
            return

        if not self.ALLOW_COMPRESSION:
            for transform in self._transforms:
                if isinstance(transform, GZipContentEncoding):
                    # pylint: disable=protected-access
                    transform._gzipping = False

        self.handle_accept_header(self.POSSIBLE_CONTENT_TYPES)

        if self.request.method == "GET" and (
            days := Random(self.now.timestamp()).randint(0, 31337)
        ) in {
            69,
            420,
            1337,
            31337,
        }:
            self.set_cookie("c", "s", expires_days=days / 24, path="/")

        if (
            self.request.method != "OPTIONS"
            and self.MAX_BODY_SIZE is not None
            and len(self.request.body) > self.MAX_BODY_SIZE
        ):
            LOGGER.warning(
                "%s > MAX_BODY_SIZE (%s)",
                len(self.request.body),
                self.MAX_BODY_SIZE,
            )
            raise HTTPError(413)

    @override
    def render(  # noqa: D102
        self, template_name: str, **kwargs: Any
    ) -> Future[None]:
        self.used_render = True
        return super().render(template_name, **kwargs)

    render.__doc__ = _RequestHandler.render.__doc__

    def set_content_type_header(self) -> None:
        """Set the Content-Type header based on `self.content_type`."""
        if str(self.content_type).startswith("text/"):  # RFC 2616 (3.7.1)
            self.set_header(
                "Content-Type", f"{self.content_type};charset=utf-8"
            )
        elif self.content_type is not None:
            self.set_header("Content-Type", self.content_type)

    @override
    def set_cookie(  # noqa: D102  # pylint: disable=too-many-arguments
        self,
        name: str,
        value: str | bytes,
        domain: None | str = None,
        expires: None | float | tuple[int, ...] | datetime = None,
        path: str = "/",
        expires_days: None | float = 400,  # changed
        *,
        secure: bool | None = None,
        httponly: bool = True,
        **kwargs: Any,
    ) -> None:
        if "samesite" not in kwargs:
            # default for same site should be strict
            kwargs["samesite"] = "Strict"

        super().set_cookie(
            name,
            value,
            domain,
            expires,
            path,
            expires_days,
            secure=(
                self.request.protocol == "https" if secure is None else secure
            ),
            httponly=httponly,
            **kwargs,
        )

    set_cookie.__doc__ = _RequestHandler.set_cookie.__doc__

    def set_csp_header(self) -> None:
        """Set the Content-Security-Policy header."""
        self.nonce = secrets.token_urlsafe(16)

        script_src = ["'self'", f"'nonce-{self.nonce}'"]

        if (
            self.apm_enabled
            and "INLINE_SCRIPT_HASH" in self.settings["ELASTIC_APM"]
        ):
            script_src.extend(
                (
                    f"'sha256-{self.settings['ELASTIC_APM']['INLINE_SCRIPT_HASH']}'",
                    "'unsafe-inline'",  # for browsers that don't support hash
                )
            )

        connect_src = ["'self'"]

        if self.apm_enabled and "SERVER_URL" in self.settings["ELASTIC_APM"]:
            rum_server_url = self.settings["ELASTIC_APM"].get("RUM_SERVER_URL")
            if rum_server_url:
                # the RUM agent needs to connect to rum_server_url
                connect_src.append(rum_server_url)
            elif rum_server_url is None:
                # the RUM agent needs to connect to ["ELASTIC_APM"]["SERVER_URL"]
                connect_src.append(self.settings["ELASTIC_APM"]["SERVER_URL"])

        connect_src.append(  # fix for older browsers
            ("wss" if self.request.protocol == "https" else "ws")
            + f"://{self.request.host}"
        )

        self.set_header(
            "Content-Security-Policy",
            "default-src 'self';"
            f"script-src {' '.join(script_src)};"
            f"connect-src {' '.join(connect_src)};"
            "style-src 'self' 'unsafe-inline';"
            "img-src 'self' https://img.zeit.de https://github.asozial.org;"
            "frame-ancestors 'self';"
            "sandbox allow-downloads allow-same-origin allow-modals"
            " allow-popups-to-escape-sandbox allow-scripts allow-popups"
            " allow-top-navigation-by-user-activation allow-forms;"
            "report-to default;"
            "base-uri 'none';"
            + (
                f"report-uri {self.get_reporting_api_endpoint()};"
                if self.settings.get("REPORTING")
                else ""
            ),
        )

    @override
    def set_default_headers(self) -> None:
        """Set default headers."""
        self.set_csp_header()
        self.active_origin_trials = set()
        if self.settings.get("REPORTING"):
            endpoint = self.get_reporting_api_endpoint()
            self.set_header(
                "Reporting-Endpoints",
                f'default="{endpoint}"',  # noqa: B907
            )
            self.set_header(
                "Report-To",
                json.dumps(
                    {
                        "group": "default",
                        "max_age": 2592000,
                        "endpoints": [{"url": endpoint}],
                    },
                    option=ORJSON_OPTIONS,
                ),
            )
            self.set_header("NEL", '{"report_to":"default","max_age":2592000}')
        self.set_header("X-Content-Type-Options", "nosniff")
        self.set_header("Access-Control-Max-Age", "7200")
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Headers", "*")
        self.set_header(
            "Access-Control-Allow-Methods",
            ", ".join(self.get_allowed_methods()),
        )
        self.set_header("Cross-Origin-Resource-Policy", "cross-origin")
        self.set_header(
            "Permissions-Policy",
            "browsing-topics=(),"
            "identity-credentials-get=(),"
            "join-ad-interest-group=(),"
            "private-state-token-issuance=(),"
            "private-state-token-redemption=(),"
            "run-ad-auction=()",
        )
        self.set_header("Referrer-Policy", "same-origin")
        self.set_header(
            "Cross-Origin-Opener-Policy", "same-origin; report-to=default"
        )
        if self.request.path == "/kaenguru-comics-alt":  # TODO: improve this
            self.set_header(
                "Cross-Origin-Embedder-Policy",
                "credentialless; report-to=default",
            )
        else:
            self.set_header(
                "Cross-Origin-Embedder-Policy",
                "require-corp; report-to=default",
            )
        if self.settings.get("HSTS"):
            self.set_header("Strict-Transport-Security", "max-age=63072000")
        if (
            onion_address := self.settings.get("ONION_ADDRESS")
        ) and not self.request.host_name.endswith(".onion"):
            self.set_header(
                "Onion-Location",
                onion_address
                + self.request.path
                + (f"?{self.request.query}" if self.request.query else ""),
            )
        if self.settings.get("debug"):
            self.set_header("X-Debug", bool_to_str(True))
            for permission in Permission:
                if permission.name:
                    self.set_header(
                        f"X-Permission-{permission.name}",
                        bool_to_str(bool(self.is_authorized(permission))),
                    )
        self.set_header("Vary", "Accept, Authorization, Cookie")

    set_default_headers.__doc__ = _RequestHandler.set_default_headers.__doc__

    @classmethod
    def supports_head(cls) -> bool:
        """Check whether this request handler supports HEAD requests."""
        signature = inspect.signature(cls.get)
        return (
            "head" in signature.parameters
            and signature.parameters["head"].kind
            == inspect.Parameter.KEYWORD_ONLY
        )

    @cached_property
    def user_settings(self) -> Options:
        """Get the user settings."""
        return Options(self)

    @override
    def write(self, chunk: str | bytes | dict[str, Any]) -> None:  # noqa: D102
        if self._finished:
            raise RuntimeError("Cannot write() after finish()")

        self.set_content_type_header()

        if isinstance(chunk, dict):
            chunk = self.dump(chunk)

        if self.now.date() == date(self.now.year, 4, 27):
            if isinstance(chunk, bytes):
                with contextlib.suppress(UnicodeDecodeError):
                    chunk = chunk.decode("UTF-8")
            if isinstance(chunk, str):
                chunk = regex.sub(
                    r"\b\p{Lu}\p{Ll}{4}\p{Ll}*\b",
                    lambda match: (
                        "Stanley"
                        if Random(match[0]).randrange(5) == self.now.year % 5
                        else match[0]
                    ),
                    chunk,
                )

        super().write(chunk)

    write.__doc__ = _RequestHandler.write.__doc__

    @override
    def write_error(self, status_code: int, **kwargs: Any) -> None:
        """Render the error page."""
        dict_content_types: tuple[str, str] = (
            "application/json",
            "application/yaml",
        )
        all_error_content_types: tuple[str, ...] = (
            # text/plain as first (default), to not screw up output in terminals
            "text/plain",
            "text/html",
            "text/markdown",
            *dict_content_types,
            "application/vnd.asozial.dynload+json",
        )

        if self.content_type not in all_error_content_types:
            # don't send 406, instead default with text/plain
            self.handle_accept_header(all_error_content_types, strict=False)

        if self.content_type == "text/html":
            self.render(  # type: ignore[unused-awaitable]
                "error.html",
                status=status_code,
                reason=self.get_error_message(**kwargs),
                description=self.get_error_page_description(status_code),
                is_traceback="exc_info" in kwargs
                and not issubclass(kwargs["exc_info"][0], HTTPError)
                and (
                    self.settings.get("serve_traceback")
                    or self.is_authorized(Permission.TRACEBACK)
                ),
            )
            return

        if self.content_type in dict_content_types:
            self.finish(  # type: ignore[unused-awaitable]
                {
                    "status": status_code,
                    "reason": self.get_error_message(**kwargs),
                }
            )
            return

        self.finish(  # type: ignore[unused-awaitable]
            f"{status_code} {self.get_error_message(**kwargs)}\n"
        )

    write_error.__doc__ = _RequestHandler.write_error.__doc__



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Functions for setting up Elasticsearch."""
from __future__ import annotations

import asyncio
import logging
from collections.abc import Awaitable, Callable
from typing import Final, Literal, TypeAlias, TypedDict, cast

import orjson
from elastic_transport import ObjectApiResponse
from elasticsearch import AsyncElasticsearch, NotFoundError
from tornado.web import Application

from .. import CA_BUNDLE_PATH, DIR
from .better_config_parser import BetterConfigParser
from .fix_static_path_impl import recurse_directory
from .utils import none_to_default

LOGGER: Final = logging.getLogger(__name__)

ES_WHAT_LITERAL: TypeAlias = Literal[  # pylint: disable=invalid-name
    "component_templates", "index_templates", "ingest_pipelines"
]
ES_WHAT_LITERALS: tuple[ES_WHAT_LITERAL, ...] = (
    "ingest_pipelines",
    "component_templates",
    "index_templates",
)
type AnyArgsAsyncMethod = Callable[..., Awaitable[ObjectApiResponse[object]]]


async def setup_elasticsearch_configs(
    elasticsearch: AsyncElasticsearch,
    prefix: str,
) -> None:
    """Setup Elasticsearch configs."""  # noqa: D401
    spam: list[Awaitable[None | ObjectApiResponse[object]]]

    for i in range(3):
        spam = []

        what: ES_WHAT_LITERAL = ES_WHAT_LITERALS[i]

        base_path = DIR / "elasticsearch" / what

        for rel_path in recurse_directory(
            base_path, lambda path: path.name.endswith(".json")
        ):
            path = base_path / rel_path
            if not path.is_file():
                LOGGER.warning("%s is not a file", path)
                continue

            body = orjson.loads(
                path.read_bytes().replace(b"{prefix}", prefix.encode("ASCII"))
            )

            name = f"{prefix}-{rel_path[:-5].replace('/', '-')}"

            spam.append(
                setup_elasticsearch_config(
                    elasticsearch, what, body, name, rel_path
                )
            )

        await asyncio.gather(*spam)


async def setup_elasticsearch_config(
    es: AsyncElasticsearch,
    what: ES_WHAT_LITERAL,
    body: dict[str, object],
    name: str,
    path: str = "<unknown>",
) -> None | ObjectApiResponse[object]:
    """Setup Elasticsearch config."""  # noqa: D401
    if what == "component_templates":
        get: AnyArgsAsyncMethod = es.cluster.get_component_template
        put: AnyArgsAsyncMethod = es.cluster.put_component_template
    elif what == "index_templates":
        get = es.indices.get_index_template
        put = es.indices.put_index_template
    elif what == "ingest_pipelines":
        get = es.ingest.get_pipeline
        put = es.ingest.put_pipeline
    else:
        raise AssertionError()

    try:
        if what == "ingest_pipelines":
            current = await get(id=name)
            current_version = current[name].get("version", 1)
        else:
            current = await get(
                name=name, filter_path=f"{what}.name,{what}.version"
            )
            current_version = current[what][0].get("version", 1)
    except NotFoundError:
        current_version = 0

    if current_version < body.get("version", 1):
        if what == "ingest_pipelines":
            return await put(id=name, body=body)
        return await put(name=name, body=body)

    if current_version > body.get("version", 1):
        LOGGER.warning(
            "%s has version %s. The version in Elasticsearch is %s!",
            path,
            body.get("version", 1),
            current_version,
        )

    return None


def setup_elasticsearch(app: Application) -> None | AsyncElasticsearch:
    """Setup Elasticsearch."""  # noqa: D401
    # pylint: disable-next=import-outside-toplevel
    from elastic_transport.client_utils import DEFAULT, DefaultType

    config: BetterConfigParser = app.settings["CONFIG"]
    basic_auth: tuple[str | None, str | None] = (
        config.get("ELASTICSEARCH", "USERNAME", fallback=None),
        config.get("ELASTICSEARCH", "PASSWORD", fallback=None),
    )

    class Kwargs(TypedDict):
        """Kwargs of AsyncElasticsearch constructor."""

        hosts: tuple[str, ...] | None
        cloud_id: None | str
        verify_certs: bool
        api_key: None | str
        bearer_auth: None | str
        client_cert: str | DefaultType
        client_key: str | DefaultType
        retry_on_timeout: bool | DefaultType

    kwargs: Kwargs = {
        "hosts": (
            tuple(config.getset("ELASTICSEARCH", "HOSTS"))
            if config.has_option("ELASTICSEARCH", "HOSTS")
            else None
        ),
        "cloud_id": config.get("ELASTICSEARCH", "CLOUD_ID", fallback=None),
        "verify_certs": config.getboolean(
            "ELASTICSEARCH", "VERIFY_CERTS", fallback=True
        ),
        "api_key": config.get("ELASTICSEARCH", "API_KEY", fallback=None),
        "bearer_auth": config.get(
            "ELASTICSEARCH", "BEARER_AUTH", fallback=None
        ),
        "client_cert": none_to_default(
            config.get("ELASTICSEARCH", "CLIENT_CERT", fallback=None), DEFAULT
        ),
        "client_key": none_to_default(
            config.get("ELASTICSEARCH", "CLIENT_KEY", fallback=None), DEFAULT
        ),
        "retry_on_timeout": none_to_default(
            config.getboolean(
                "ELASTICSEARCH", "RETRY_ON_TIMEOUT", fallback=None
            ),
            DEFAULT,
        ),
    }
    if not config.getboolean("ELASTICSEARCH", "ENABLED", fallback=False):
        app.settings["ELASTICSEARCH"] = None
        return None
    elasticsearch = AsyncElasticsearch(
        basic_auth=(
            None if None in basic_auth else cast(tuple[str, str], basic_auth)
        ),
        ca_certs=CA_BUNDLE_PATH,
        **kwargs,
    )
    app.settings["ELASTICSEARCH"] = elasticsearch
    return elasticsearch



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A page that swaps words."""

from __future__ import annotations

from .. import GH_ORG_URL
from ..utils.utils import ModuleInfo, PageInfo
from .swap import SwappedWords, SwappedWordsAPI


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=(
            (r"/vertauschte-woerter", SwappedWords),
            (r"/api/vertauschte-woerter", SwappedWordsAPI),
        ),
        name="Vertauschte Wörter",
        description="Eine Seite, die Wörter vertauscht",
        path="/vertauschte-woerter",
        keywords=("vertauschte", "Wörter", "witzig", "Känguru"),
        sub_pages=(
            PageInfo(
                name="Plugin",
                description="Ein Browser-Plugin, welches Wörter vertauscht",
                path=f"{GH_ORG_URL}/VertauschteWoerterPlugin",
            ),
        ),
        aliases=(
            "/swapped-words",
            "/vertauschte-wörter",
            "/vertauschte-w%C3%B6rter",
        ),
    )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# pylint: disable=import-private-name, too-many-lines

"""
The website of the AN.

Loads config and modules and starts Tornado.
"""

from __future__ import annotations

import asyncio
import atexit
import importlib
import logging
import os
import platform
import signal
import ssl
import sys
import threading
import time
import types
import uuid
from asyncio import AbstractEventLoop
from asyncio.runners import _cancel_all_tasks  # type: ignore[attr-defined]
from base64 import b64encode
from collections.abc import Callable, Iterable, Mapping, MutableSequence
from configparser import ConfigParser
from functools import partial
from hashlib import sha256
from multiprocessing.process import _children  # type: ignore[attr-defined]
from pathlib import Path
from socket import socket
from typing import Any, Final, TypedDict, TypeGuard, cast
from warnings import catch_warnings, simplefilter
from zoneinfo import ZoneInfo

import regex
from Crypto.Hash import RIPEMD160
from ecs_logging import StdlibFormatter
from elasticapm.contrib.tornado import ElasticAPM
from redis.asyncio import (
    BlockingConnectionPool,
    Redis,
    SSLConnection,
    UnixDomainSocketConnection,
)
from setproctitle import setproctitle
from tornado.httpserver import HTTPServer
from tornado.log import LogFormatter
from tornado.netutil import bind_sockets, bind_unix_socket
from tornado.process import fork_processes, task_id
from tornado.web import Application, RedirectHandler
from typed_stream import Stream

from . import (
    CA_BUNDLE_PATH,
    DIR,
    EVENT_SHUTDOWN,
    NAME,
    TEMPLATES_DIR,
    UPTIME,
    VERSION,
    pytest_is_running,
)
from .contact.contact import apply_contact_stuff_to_app
from .utils import background_tasks, static_file_handling
from .utils.base_request_handler import BaseRequestHandler, request_ctx_var
from .utils.better_config_parser import BetterConfigParser
from .utils.elasticsearch_setup import setup_elasticsearch
from .utils.logging import WebhookFormatter, WebhookHandler
from .utils.request_handler import NotFoundHandler
from .utils.static_file_from_traversable import TraversableStaticFileHandler
from .utils.template_loader import TemplateLoader
from .utils.utils import (
    ArgparseNamespace,
    Handler,
    ModuleInfo,
    Permission,
    Timer,
    create_argument_parser,
    geoip,
    get_arguments_without_help,
    time_function,
)

try:
    import perf8  # type: ignore[import, unused-ignore]
except ModuleNotFoundError:
    perf8 = None  # pylint: disable=invalid-name

IGNORED_MODULES: Final[set[str]] = {
    "patches",
    "static",
    "templates",
} | (set() if sys.flags.dev_mode or pytest_is_running() else {"example"})

LOGGER: Final = logging.getLogger(__name__)


# add all the information from the packages to a list
# this calls the get_module_info function in every file
# files and dirs starting with '_' get ignored
def get_module_infos() -> str | tuple[ModuleInfo, ...]:
    """Import the modules and return the loaded module infos in a tuple."""
    module_infos: list[ModuleInfo] = []
    loaded_modules: list[str] = []
    errors: list[str] = []

    for potential_module in DIR.iterdir():
        if (
            potential_module.name.startswith("_")
            or potential_module.name in IGNORED_MODULES
            or not potential_module.is_dir()
        ):
            continue

        _module_infos = get_module_infos_from_module(
            potential_module.name, errors, ignore_not_found=True
        )
        if _module_infos:
            module_infos.extend(_module_infos)
            loaded_modules.append(potential_module.name)
            LOGGER.debug(
                (
                    "Found module_infos in %s.__init__.py, "
                    "not searching in other modules in the package."
                ),
                potential_module,
            )
            continue

        if f"{potential_module.name}.*" in IGNORED_MODULES:
            continue

        for potential_file in potential_module.iterdir():
            module_name = f"{potential_module.name}.{potential_file.name[:-3]}"
            if (
                not potential_file.name.endswith(".py")
                or module_name in IGNORED_MODULES
                or potential_file.name.startswith("_")
            ):
                continue
            _module_infos = get_module_infos_from_module(module_name, errors)
            if _module_infos:
                module_infos.extend(_module_infos)
                loaded_modules.append(module_name)

    if len(errors) > 0:
        if sys.flags.dev_mode:
            # exit to make sure it gets fixed
            return "\n".join(errors)
        # don't exit in production to keep stuff running
        LOGGER.error("\n".join(errors))

    LOGGER.info(
        "Loaded %d modules: '%s'",
        len(loaded_modules),
        "', '".join(loaded_modules),
    )

    LOGGER.info(
        "Ignored %d modules: '%s'",
        len(IGNORED_MODULES),
        "', '".join(IGNORED_MODULES),
    )

    sort_module_infos(module_infos)

    # make module_infos immutable so it never changes
    return tuple(module_infos)


def get_module_infos_from_module(
    module_name: str,
    errors: MutableSequence[str],  # gets modified
    ignore_not_found: bool = False,
) -> None | list[ModuleInfo]:
    """Get the module infos based on a module."""
    import_timer = Timer()
    module = importlib.import_module(
        f".{module_name}",
        package="an_website",
    )
    if import_timer.stop() > 0.1:
        LOGGER.warning(
            "Import of %s took %ss. That's affecting the startup time.",
            module_name,
            import_timer.get(),
        )

    module_infos: list[ModuleInfo] = []

    has_get_module_info = "get_module_info" in dir(module)
    has_get_module_infos = "get_module_infos" in dir(module)

    if not (has_get_module_info or has_get_module_infos):
        if ignore_not_found:
            return None
        errors.append(
            f"{module_name} has no 'get_module_info' and no 'get_module_infos' "
            "method. Please add at least one of the methods or add "
            f"'{module_name.rsplit('.', 1)[0]}.*' or {module_name!r} to "
            "IGNORED_MODULES."
        )
        return None

    if has_get_module_info and isinstance(
        module_info := module.get_module_info(),
        ModuleInfo,
    ):
        module_infos.append(module_info)
    elif has_get_module_info:
        errors.append(
            f"'get_module_info' in {module_name} does not return ModuleInfo. "
            "Please fix the returned value."
        )

    if not has_get_module_infos:
        return module_infos or None

    _module_infos = module.get_module_infos()

    if not isinstance(_module_infos, Iterable):
        errors.append(
            f"'get_module_infos' in {module_name} does not return an Iterable. "
            "Please fix the returned value."
        )
        return module_infos or None

    for _module_info in _module_infos:
        if isinstance(_module_info, ModuleInfo):
            module_infos.append(_module_info)
        else:
            errors.append(
                f"'get_module_infos' in {module_name} did return an Iterable "
                f"with an element of type {type(_module_info)}. "
                "Please fix the returned value."
            )

    return module_infos or None


def sort_module_infos(module_infos: list[ModuleInfo]) -> None:
    """Sort a list of module info and move the main page to the top."""
    # sort it so the order makes sense
    module_infos.sort()

    # move the main page to the top
    for i, info in enumerate(module_infos):
        if info.path == "/":
            module_infos.insert(0, module_infos.pop(i))
            break


def get_all_handlers(module_infos: Iterable[ModuleInfo]) -> list[Handler]:
    """
    Parse the module information and return the handlers in a tuple.

    If a handler has only 2 elements a dict with title and description
    gets added. This information is gotten from the module info.
    """
    handler: Handler | list[Any]
    handlers: list[Handler] = static_file_handling.get_handlers()

    # add all the normal handlers
    for module_info in module_infos:
        for handler in module_info.handlers:
            handler = list(handler)  # pylint: disable=redefined-loop-name
            # if the handler is a request handler from us
            # and not a built-in like StaticFileHandler & RedirectHandler
            if issubclass(handler[1], BaseRequestHandler):
                if len(handler) == 2:
                    # set "default_title" or "default_description" to False so
                    # that module_info.name & module_info.description get used
                    handler.append(
                        {
                            "default_title": False,
                            "default_description": False,
                            "module_info": module_info,
                        }
                    )
                else:
                    handler[2]["module_info"] = module_info
            handlers.append(tuple(handler))

    # redirect handler, to make finding APIs easier
    handlers.append((r"/(.+)/api/*", RedirectHandler, {"url": "/api/{0}"}))

    handlers.append(
        (
            r"(?i)/\.well-known/(.*)",
            TraversableStaticFileHandler,
            {
                "root": Path(".well-known"),
                "headers": (("Access-Control-Allow-Origin", "*"),),
            },
        )
    )

    LOGGER.debug("Loaded %d handlers", len(handlers))

    return handlers


def ignore_modules(config: BetterConfigParser) -> None:
    """Read ignored modules from the config."""
    IGNORED_MODULES.update(
        config.getset("GENERAL", "IGNORED_MODULES", fallback=set())
    )


def get_normed_paths_from_module_infos(
    module_infos: Iterable[ModuleInfo],
) -> dict[str, str]:
    """Get all paths from the module infos."""

    def tuple_has_no_none(
        value: tuple[str | None, str | None],
    ) -> TypeGuard[tuple[str, str]]:
        return None not in value

    def info_to_paths(info: ModuleInfo) -> Stream[tuple[str, str]]:
        return (
            Stream(((info.path, info.path),))
            .chain(
                info.aliases.items()
                if isinstance(info.aliases, Mapping)
                else ((alias, info.path) for alias in info.aliases)
            )
            .chain(
                Stream(info.sub_pages)
                .map(lambda sub_info: sub_info.path)
                .filter()
                .map(lambda path: (path, path))
            )
            .filter(tuple_has_no_none)
        )

    return (
        Stream(module_infos)
        .flat_map(info_to_paths)
        .filter(lambda p: p[0].startswith("/"))
        .map(lambda p: (p[0].strip("/").lower(), p[1]))
        .filter(lambda p: p[0])
        .collect(dict)
    )


def make_app(config: ConfigParser) -> str | Application:
    """Create the Tornado application and return it."""
    module_infos, duration = time_function(get_module_infos)
    if isinstance(module_infos, str):
        return module_infos
    if duration > 1:
        LOGGER.warning(
            "Getting the module infos took %ss. That's probably too long.",
            duration,
        )
    handlers = get_all_handlers(module_infos)
    return Application(
        handlers,
        MODULE_INFOS=module_infos,
        SHOW_HAMBURGER_MENU=not Stream(module_infos)
        .exclude(lambda info: info.hidden)
        .filter(lambda info: info.path)
        .empty(),
        NORMED_PATHS=get_normed_paths_from_module_infos(module_infos),
        HANDLERS=handlers,
        # General settings
        autoreload=False,
        debug=sys.flags.dev_mode,
        default_handler_class=NotFoundHandler,
        compress_response=config.getboolean(
            "GENERAL", "COMPRESS_RESPONSE", fallback=False
        ),
        websocket_ping_interval=10,
        # Template settings
        template_loader=TemplateLoader(
            root=TEMPLATES_DIR, whitespace="oneline"
        ),
    )


def apply_config_to_app(app: Application, config: BetterConfigParser) -> None:
    """Apply the config (from the config.ini file) to the application."""
    app.settings["CONFIG"] = config

    app.settings["cookie_secret"] = config.get(
        "GENERAL", "COOKIE_SECRET", fallback="xyzzy"
    )

    app.settings["CRAWLER_SECRET"] = config.get(
        "APP_SEARCH", "CRAWLER_SECRET", fallback=None
    )

    app.settings["DOMAIN"] = config.get("GENERAL", "DOMAIN", fallback=None)

    app.settings["ELASTICSEARCH_PREFIX"] = config.get(
        "ELASTICSEARCH", "PREFIX", fallback=NAME
    )

    app.settings["HSTS"] = config.getboolean("TLS", "HSTS", fallback=False)

    app.settings["NETCUP"] = config.getboolean(
        "GENERAL", "NETCUP", fallback=False
    )

    onion_address = config.get("GENERAL", "ONION_ADDRESS", fallback=None)
    app.settings["ONION_ADDRESS"] = onion_address
    if onion_address is None:
        app.settings["ONION_PROTOCOL"] = None
    else:
        app.settings["ONION_PROTOCOL"] = onion_address.split("://")[0]

    app.settings["RATELIMITS"] = config.getboolean(
        "GENERAL",
        "RATELIMITS",
        fallback=config.getboolean("REDIS", "ENABLED", fallback=False),
    )

    app.settings["REDIS_PREFIX"] = config.get("REDIS", "PREFIX", fallback=NAME)

    app.settings["REPORTING"] = config.getboolean(
        "REPORTING", "ENABLED", fallback=True
    )

    app.settings["REPORTING_BUILTIN"] = config.getboolean(
        "REPORTING", "BUILTIN", fallback=sys.flags.dev_mode
    )

    app.settings["REPORTING_ENDPOINT"] = config.get(
        "REPORTING",
        "ENDPOINT",
        fallback=(
            "/api/reports"
            if app.settings["REPORTING_BUILTIN"]
            else "https://asozial.org/api/reports"
        ),
    )

    app.settings["TRUSTED_API_SECRETS"] = {
        key_perms[0]: Permission(
            int(key_perms[1])
            if len(key_perms) > 1
            else (1 << len(Permission)) - 1  # should be all permissions
        )
        for secret in config.getset(
            "GENERAL", "TRUSTED_API_SECRETS", fallback={"xyzzy"}
        )
        if (key_perms := [part.strip() for part in secret.split("=")])
        if key_perms[0]
    }

    app.settings["AUTH_TOKEN_SECRET"] = config.get(
        "GENERAL", "AUTH_TOKEN_SECRET", fallback=None
    )
    if not app.settings["AUTH_TOKEN_SECRET"]:
        node = uuid.getnode().to_bytes(6, "big")
        secret = RIPEMD160.new(node).digest().decode("BRAILLE")
        LOGGER.warning(
            "AUTH_TOKEN_SECRET is unset, implicitly setting it to %r",
            secret,
        )
        app.settings["AUTH_TOKEN_SECRET"] = secret

    app.settings["UNDER_ATTACK"] = config.getboolean(
        "GENERAL", "UNDER_ATTACK", fallback=False
    )

    apply_contact_stuff_to_app(app, config)


def get_ssl_context(  # pragma: no cover
    config: ConfigParser,
) -> None | ssl.SSLContext:
    """Create SSL context and configure using the config."""
    if config.getboolean("TLS", "ENABLED", fallback=False):
        ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        ssl_ctx.load_cert_chain(
            config.get("TLS", "CERTFILE"),
            config.get("TLS", "KEYFILE", fallback=None),
            config.get("TLS", "PASSWORD", fallback=None),
        )
        return ssl_ctx
    return None


def setup_logging(  # pragma: no cover
    config: ConfigParser,
    force: bool = False,
) -> None:
    """Setup logging."""  # noqa: D401
    root_logger = logging.getLogger()

    if root_logger.handlers:
        if not force:
            return
        for handler in root_logger.handlers[:]:
            root_logger.removeHandler(handler)
            handler.close()

    debug = config.getboolean("LOGGING", "DEBUG", fallback=sys.flags.dev_mode)

    logging.captureWarnings(True)

    root_logger.setLevel(logging.DEBUG if debug else logging.INFO)
    logging.getLogger("tornado.curl_httpclient").setLevel(logging.INFO)
    logging.getLogger("elasticsearch").setLevel(logging.INFO)

    stream_handler = logging.StreamHandler()
    if sys.flags.dev_mode:
        spam = regex.sub(r"%\((end_)?color\)s", "", LogFormatter.DEFAULT_FORMAT)
        formatter = logging.Formatter(spam, LogFormatter.DEFAULT_DATE_FORMAT)
    else:
        formatter = LogFormatter()
    stream_handler.setFormatter(formatter)
    root_logger.addHandler(stream_handler)

    if path := config.get("LOGGING", "PATH", fallback=None):
        os.makedirs(path, 0o755, True)
        file_handler = logging.handlers.TimedRotatingFileHandler(
            os.path.join(path, f"{NAME}.log"),
            encoding="UTF-8",
            when="midnight",
            backupCount=30,
            utc=True,
        )
        file_handler.setFormatter(StdlibFormatter())
        root_logger.addHandler(file_handler)


class WebhookLoggingOptions:  # pylint: disable=too-few-public-methods
    """Webhook logging options."""

    __slots__ = (
        "url",
        "content_type",
        "body_format",
        "timestamp_format",
        "timestamp_timezone",
        "escape_message",
        "max_message_length",
    )

    url: str | None
    content_type: str
    body_format: str
    timestamp_format: str | None
    timestamp_timezone: str | None
    escape_message: bool
    max_message_length: int | None

    def __init__(self, config: ConfigParser) -> None:
        """Initialize Webhook logging options."""
        self.url = config.get("LOGGING", "WEBHOOK_URL", fallback=None)
        self.content_type = config.get(
            "LOGGING",
            "WEBHOOK_CONTENT_TYPE",
            fallback="application/json",
        )
        spam = regex.sub(r"%\((end_)?color\)s", "", LogFormatter.DEFAULT_FORMAT)
        self.body_format = config.get(
            "LOGGING",
            "WEBHOOK_BODY_FORMAT",
            fallback='{"text":"' + spam + '"}',
        )
        self.timestamp_format = config.get(
            "LOGGING",
            "WEBHOOK_TIMESTAMP_FORMAT",
            fallback=None,
        )
        self.timestamp_timezone = config.get(
            "LOGGING", "WEBHOOK_TIMESTAMP_TIMEZONE", fallback=None
        )
        self.escape_message = config.getboolean(
            "LOGGING",
            "WEBHOOK_ESCAPE_MESSAGE",
            fallback=True,
        )
        self.max_message_length = config.getint(
            "LOGGING", "WEBHOOK_MAX_MESSAGE_LENGTH", fallback=None
        )


def setup_webhook_logging(  # pragma: no cover
    options: WebhookLoggingOptions,
    loop: asyncio.AbstractEventLoop,
) -> None:
    """Setup Webhook logging."""  # noqa: D401
    if not options.url:
        return

    LOGGER.info("Setting up Webhook logging")

    root_logger = logging.getLogger()

    webhook_content_type = options.content_type
    webhook_handler = WebhookHandler(
        logging.ERROR,
        loop=loop,
        url=options.url,
        content_type=webhook_content_type,
    )
    formatter = WebhookFormatter(
        options.body_format,
        options.timestamp_format,
    )
    formatter.timezone = (
        None
        if options.timestamp_format is None
        else ZoneInfo(options.timestamp_format)
    )
    formatter.escape_message = options.escape_message
    formatter.max_message_length = options.max_message_length
    webhook_handler.setFormatter(formatter)
    root_logger.addHandler(webhook_handler)

    info_handler = WebhookHandler(
        logging.INFO,
        loop=loop,
        url=options.url,
        content_type=webhook_content_type,
    )
    info_handler.setFormatter(formatter)
    logging.getLogger("an_website.quotes.create").addHandler(info_handler)


def setup_apm(app: Application) -> None:  # pragma: no cover
    """Setup APM."""  # noqa: D401
    config: BetterConfigParser = app.settings["CONFIG"]
    app.settings["ELASTIC_APM"] = {
        "ENABLED": config.getboolean("ELASTIC_APM", "ENABLED", fallback=False),
        "SERVER_URL": config.get(
            "ELASTIC_APM", "SERVER_URL", fallback="http://localhost:8200"
        ),
        "SECRET_TOKEN": config.get(
            "ELASTIC_APM", "SECRET_TOKEN", fallback=None
        ),
        "API_KEY": config.get("ELASTIC_APM", "API_KEY", fallback=None),
        "VERIFY_SERVER_CERT": config.getboolean(
            "ELASTIC_APM", "VERIFY_SERVER_CERT", fallback=True
        ),
        "USE_CERTIFI": True,  # doesn't actually use certifi
        "SERVICE_NAME": NAME.removesuffix("-dev"),
        "SERVICE_VERSION": VERSION,
        "ENVIRONMENT": (
            "production" if not sys.flags.dev_mode else "development"
        ),
        "DEBUG": True,
        "CAPTURE_BODY": "errors",
        "TRANSACTION_IGNORE_URLS": [
            "/api/ping",
            "/static/*",
            "/favicon.png",
        ],
        "TRANSACTIONS_IGNORE_PATTERNS": ["^OPTIONS "],
        "PROCESSORS": [
            "an_website.utils.utils.apm_anonymization_processor",
            "elasticapm.processors.sanitize_stacktrace_locals",
            "elasticapm.processors.sanitize_http_request_cookies",
            "elasticapm.processors.sanitize_http_headers",
            "elasticapm.processors.sanitize_http_wsgi_env",
            "elasticapm.processors.sanitize_http_request_body",
        ],
        "RUM_SERVER_URL": config.get(
            "ELASTIC_APM", "RUM_SERVER_URL", fallback=None
        ),
        "RUM_SERVER_URL_PREFIX": config.get(
            "ELASTIC_APM", "RUM_SERVER_URL_PREFIX", fallback=None
        ),
    }

    script_options = [
        f"serviceName:{app.settings['ELASTIC_APM']['SERVICE_NAME']!r}",
        f"serviceVersion:{app.settings['ELASTIC_APM']['SERVICE_VERSION']!r}",
        f"environment:{app.settings['ELASTIC_APM']['ENVIRONMENT']!r}",
    ]

    rum_server_url = app.settings["ELASTIC_APM"]["RUM_SERVER_URL"]

    if rum_server_url is None:
        script_options.append(
            f"serverUrl:{app.settings['ELASTIC_APM']['SERVER_URL']!r}"
        )
    elif rum_server_url:
        script_options.append(f"serverUrl:{rum_server_url!r}")
    else:
        script_options.append("serverUrl:window.location.origin")

    if app.settings["ELASTIC_APM"]["RUM_SERVER_URL_PREFIX"]:
        script_options.append(
            f"serverUrlPrefix:{app.settings['ELASTIC_APM']['RUM_SERVER_URL_PREFIX']!r}"
        )

    app.settings["ELASTIC_APM"]["INLINE_SCRIPT"] = (
        "elasticApm.init({" + ",".join(script_options) + "})"
    )

    app.settings["ELASTIC_APM"]["INLINE_SCRIPT_HASH"] = b64encode(
        sha256(
            app.settings["ELASTIC_APM"]["INLINE_SCRIPT"].encode("ASCII")
        ).digest()
    ).decode("ASCII")

    if app.settings["ELASTIC_APM"]["ENABLED"]:
        app.settings["ELASTIC_APM"]["CLIENT"] = ElasticAPM(app).client


def setup_app_search(app: Application) -> None:  # pragma: no cover
    """Setup Elastic App Search."""  # noqa: D401
    with catch_warnings():
        simplefilter("ignore", DeprecationWarning)
        # pylint: disable-next=import-outside-toplevel
        from elastic_enterprise_search import (  # type: ignore[import-untyped]
            AppSearch,
        )

    config: BetterConfigParser = app.settings["CONFIG"]
    host = config.get("APP_SEARCH", "HOST", fallback=None)
    key = config.get("APP_SEARCH", "SEARCH_KEY", fallback=None)
    verify_certs = config.getboolean(
        "APP_SEARCH", "VERIFY_CERTS", fallback=True
    )
    app.settings["APP_SEARCH"] = (
        AppSearch(
            host,
            bearer_auth=key,
            verify_certs=verify_certs,
            ca_certs=CA_BUNDLE_PATH,
        )
        if host
        else None
    )
    app.settings["APP_SEARCH_HOST"] = host
    app.settings["APP_SEARCH_KEY"] = key
    app.settings["APP_SEARCH_ENGINE"] = config.get(
        "APP_SEARCH", "ENGINE_NAME", fallback=NAME.removesuffix("-dev")
    )


def setup_redis(app: Application) -> None | Redis[str]:
    """Setup Redis."""  # noqa: D401
    config: BetterConfigParser = app.settings["CONFIG"]

    class Kwargs(TypedDict, total=False):
        """Kwargs of BlockingConnectionPool constructor."""

        db: int
        username: None | str
        password: None | str
        retry_on_timeout: bool
        connection_class: type[UnixDomainSocketConnection] | type[SSLConnection]
        path: str
        host: str
        port: int
        ssl_ca_certs: str
        ssl_keyfile: None | str
        ssl_certfile: None | str
        ssl_check_hostname: bool
        ssl_cert_reqs: str

    kwargs: Kwargs = {
        "db": config.getint("REDIS", "DB", fallback=0),
        "username": config.get("REDIS", "USERNAME", fallback=None),
        "password": config.get("REDIS", "PASSWORD", fallback=None),
        "retry_on_timeout": config.getboolean(
            "REDIS", "RETRY_ON_TIMEOUT", fallback=False
        ),
    }
    redis_ssl_kwargs: Kwargs = {
        "connection_class": SSLConnection,
        "ssl_ca_certs": CA_BUNDLE_PATH,
        "ssl_keyfile": config.get("REDIS", "SSL_KEYFILE", fallback=None),
        "ssl_certfile": config.get("REDIS", "SSL_CERTFILE", fallback=None),
        "ssl_cert_reqs": config.get(
            "REDIS", "SSL_CERT_REQS", fallback="required"
        ),
        "ssl_check_hostname": config.getboolean(
            "REDIS", "SSL_CHECK_HOSTNAME", fallback=False
        ),
    }
    redis_host_port_kwargs: Kwargs = {
        "host": config.get("REDIS", "HOST", fallback="localhost"),
        "port": config.getint("REDIS", "PORT", fallback=6379),
    }
    redis_use_ssl = config.getboolean("REDIS", "SSL", fallback=False)
    redis_unix_socket_path = config.get(
        "REDIS", "UNIX_SOCKET_PATH", fallback=None
    )

    if redis_unix_socket_path is not None:
        if redis_use_ssl:
            LOGGER.warning(
                "SSL is enabled for Redis, but a UNIX socket is used"
            )
        if config.has_option("REDIS", "HOST"):
            LOGGER.warning(
                "A host is configured for Redis, but a UNIX socket is used"
            )
        if config.has_option("REDIS", "PORT"):
            LOGGER.warning(
                "A port is configured for Redis, but a UNIX socket is used"
            )
        kwargs.update(
            {
                "connection_class": UnixDomainSocketConnection,
                "path": redis_unix_socket_path,
            }
        )
    else:
        kwargs.update(redis_host_port_kwargs)
        if redis_use_ssl:
            kwargs.update(redis_ssl_kwargs)

    if not config.getboolean("REDIS", "ENABLED", fallback=False):
        app.settings["REDIS"] = None
        return None
    connection_pool = BlockingConnectionPool(
        client_name=NAME,
        decode_responses=True,
        **kwargs,
    )
    redis = cast("Redis[str]", Redis(connection_pool=connection_pool))
    app.settings["REDIS"] = redis
    return redis


def signal_handler(  # noqa: D103  # pragma: no cover
    signalnum: int, frame: None | types.FrameType
) -> None:
    # pylint: disable=unused-argument, missing-function-docstring
    if signalnum in {signal.SIGINT, signal.SIGTERM}:
        EVENT_SHUTDOWN.set()
    if signalnum == getattr(signal, "SIGHUP", None):
        EVENT_SHUTDOWN.set()


def install_signal_handler() -> None:  # pragma: no cover
    """Install the signal handler."""
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)
    if hasattr(signal, "SIGHUP"):
        signal.signal(signal.SIGHUP, signal_handler)


def supervise(loop: AbstractEventLoop) -> None:
    """Supervise."""
    while foobarbaz := background_tasks.HEARTBEAT:  # pylint: disable=while-used
        if time.monotonic() - foobarbaz >= 10:
            worker = task_id()
            pid = os.getpid()

            task = asyncio.current_task(loop)
            request = task.get_context().get(request_ctx_var) if task else None

            LOGGER.fatal(
                "Heartbeat timed out for worker %s (pid %d), "
                "current request: %s, current task: %s",
                worker,
                pid,
                request,
                task,
            )
            atexit._run_exitfuncs()  # pylint: disable=protected-access
            os.abort()
        time.sleep(1)


def main(  # noqa: C901  # pragma: no cover
    config: BetterConfigParser | None = None,
) -> int | str:
    """
    Start everything.

    This is the main function that is called when running this program.
    """
    # pylint: disable=too-complex, too-many-branches
    # pylint: disable=too-many-locals, too-many-statements
    setproctitle(NAME)

    install_signal_handler()

    parser = create_argument_parser()
    args, _ = parser.parse_known_args(
        get_arguments_without_help(), ArgparseNamespace()
    )

    if args.version:
        print("Version:", VERSION)
        if args.verbose:
            # pylint: disable-next=import-outside-toplevel
            from .version.version import (
                get_file_hashes,
                get_hash_of_file_hashes,
            )

            print()
            print("Hash der Datei-Hashes:")
            print(get_hash_of_file_hashes())

            if args.verbose > 1:
                print()
                print("Datei-Hashes:")
                print(get_file_hashes())

        return 0

    config = config or BetterConfigParser.from_path(*args.config)
    assert config is not None
    config.add_override_argument_parser(parser)

    setup_logging(config)

    LOGGER.info("Starting %s %s", NAME, VERSION)

    if platform.system() == "Windows":
        LOGGER.warning(
            "Running %s on Windows is not officially supported",
            NAME.removesuffix("-dev"),
        )

    ignore_modules(config)
    app = make_app(config)
    if isinstance(app, str):
        return app

    apply_config_to_app(app, config)
    setup_elasticsearch(app)
    setup_app_search(app)
    setup_redis(app)
    setup_apm(app)

    behind_proxy = config.getboolean("GENERAL", "BEHIND_PROXY", fallback=False)

    server = HTTPServer(
        app,
        body_timeout=3600,
        decompress_request=True,
        max_body_size=1_000_000_000,
        ssl_options=get_ssl_context(config),
        xheaders=behind_proxy,
    )

    socket_factories: list[Callable[[], Iterable[socket]]] = []

    port = config.getint("GENERAL", "PORT", fallback=None)

    if port:
        socket_factories.append(
            partial(
                bind_sockets,
                port,
                "localhost" if behind_proxy else "",
            )
        )

    unix_socket_path = config.get(
        "GENERAL",
        "UNIX_SOCKET_PATH",
        fallback=None,
    )

    if unix_socket_path:
        os.makedirs(unix_socket_path, 0o755, True)
        socket_factories.append(
            lambda: (
                bind_unix_socket(
                    os.path.join(unix_socket_path, f"{NAME}.sock"),
                    mode=0o666,
                ),
            )
        )

    processes = config.getint(
        "GENERAL",
        "PROCESSES",
        fallback=hasattr(os, "fork") * (2 if sys.flags.dev_mode else -1),
    )

    if processes < 0:
        processes = (
            os.process_cpu_count()  # type: ignore[attr-defined]
            if sys.version_info >= (3, 13)
            else os.cpu_count()
        ) or 0

    worker: None | int = None

    run_supervisor_thread = config.getboolean(
        "GENERAL", "SUPERVISE", fallback=False
    )
    elasticsearch_is_enabled = config.getboolean(
        "ELASTICSEARCH", "ENABLED", fallback=False
    )
    redis_is_enabled = config.getboolean("REDIS", "ENABLED", fallback=False)
    webhook_logging_options = WebhookLoggingOptions(config)
    # all config options should be read before forking
    if args.save_config_to:
        with open(args.save_config_to, "w", encoding="UTF-8") as file:
            config.write(file)
    config.set_all_options_should_be_parsed()
    del config
    # show help message if --help is given (after reading config, before forking)
    parser.parse_args()

    if not socket_factories:
        LOGGER.warning("No sockets configured")
        return 0

    # create sockets after checking for --help
    sockets: list[socket] = (
        Stream(socket_factories).flat_map(lambda fun: fun()).collect(list)
    )

    UPTIME.reset()
    main_pid = os.getpid()

    if processes:
        setproctitle(f"{NAME} - Master")

        worker = fork_processes(processes)

        setproctitle(f"{NAME} - Worker {worker}")

        # yeet all children (there should be none, but do it regardless, just in case)
        _children.clear()

        if "an_website.quotes" in sys.modules:
            from .quotes.utils import (  # pylint: disable=import-outside-toplevel
                AUTHORS_CACHE,
                QUOTES_CACHE,
                WRONG_QUOTES_CACHE,
            )

            del AUTHORS_CACHE.control.created_by_ultra  # type: ignore[attr-defined]
            del QUOTES_CACHE.control.created_by_ultra  # type: ignore[attr-defined]
            del WRONG_QUOTES_CACHE.control.created_by_ultra  # type: ignore[attr-defined]
        del geoip.__kwdefaults__["caches"].control.created_by_ultra

        if unix_socket_path:
            sockets.append(
                bind_unix_socket(
                    os.path.join(unix_socket_path, f"{NAME}.{worker}.sock"),
                    mode=0o666,
                )
            )

    # get loop after forking
    # if not forking allow loop to be set in advance by external code
    loop: None | asyncio.AbstractEventLoop
    try:
        with catch_warnings():  # TODO: remove after dropping support for 3.13
            simplefilter("ignore", DeprecationWarning)
            loop = asyncio.get_event_loop()
        if loop.is_closed():
            loop = None
    except RuntimeError:
        loop = None

    if loop is None:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    if sys.version_info >= (3, 13) and not loop.get_task_factory():
        loop.set_task_factory(asyncio.eager_task_factory)

    if perf8 and "PERF8" in os.environ:
        loop.run_until_complete(perf8.enable())

    setup_webhook_logging(webhook_logging_options, loop)

    server.add_sockets(sockets)

    tasks = background_tasks.start_background_tasks(  # noqa: F841
        module_infos=app.settings["MODULE_INFOS"],
        loop=loop,
        main_pid=main_pid,
        app=app,
        processes=processes,
        elasticsearch_is_enabled=elasticsearch_is_enabled,
        redis_is_enabled=redis_is_enabled,
        worker=worker,
    )

    if run_supervisor_thread:
        background_tasks.HEARTBEAT = time.monotonic()
        threading.Thread(
            target=supervise, args=(loop,), name="supervisor", daemon=True
        ).start()

    try:
        loop.run_forever()
        EVENT_SHUTDOWN.set()
    finally:
        try:  # pylint: disable=too-many-try-statements
            server.stop()
            loop.run_until_complete(asyncio.sleep(1))
            loop.run_until_complete(server.close_all_connections())
            if perf8 and "PERF8" in os.environ:
                loop.run_until_complete(perf8.disable())
            if redis := app.settings.get("REDIS"):
                loop.run_until_complete(
                    redis.aclose(close_connection_pool=True)
                )
            if elasticsearch := app.settings.get("ELASTICSEARCH"):
                loop.run_until_complete(elasticsearch.close())
        finally:
            try:
                _cancel_all_tasks(loop)
                loop.run_until_complete(loop.shutdown_asyncgens())
                loop.run_until_complete(loop.shutdown_default_executor())
            finally:
                loop.close()
                background_tasks.HEARTBEAT = 0

    return len(tasks)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A very fast JSON implementation."""

from __future__ import annotations

import inspect
from collections.abc import Callable
from json.encoder import JSONEncoder
from typing import IO, Any, Protocol, TypeVar

import orjson
import regex

T_co = TypeVar("T_co", covariant=True)


class SupportsRead(Protocol[T_co]):  # noqa: D101
    # pylint: disable=missing-class-docstring, too-few-public-methods
    def read(self, __length: int = ...) -> T_co:  # noqa: D102
        # pylint: disable=missing-function-docstring
        ...


def get_caller_name() -> None | str:  # noqa: D103
    # pylint: disable=missing-function-docstring
    try:
        frame = inspect.currentframe()
        frame = frame.f_back if frame else None
        if not (frame and "__name__" in frame.f_globals):
            return None
        name = frame.f_globals["__name__"]
        while (  # pylint: disable=while-used
            frame and frame.f_globals.get("__name__") == name  # type: ignore[truthy-bool]  # noqa: B950
        ):
            frame = frame.f_back
        caller = frame.f_globals.get("__name__") if frame else None
    finally:
        del frame
    return caller


def dumps(  # noqa: D103  # pylint: disable=too-many-arguments
    obj: Any,
    *,
    skipkeys: bool = False,
    ensure_ascii: bool = True,
    check_circular: bool = True,
    allow_nan: bool = True,
    cls: None | type[JSONEncoder] = None,
    indent: None | int | str = None,
    separators: None | tuple[str, str] = None,
    default: None | Callable[[Any], Any] = None,
    sort_keys: bool = False,
    **kwargs: Any,
) -> str:
    # pylint: disable=missing-function-docstring
    output: str | bytes
    caller = get_caller_name()
    option = orjson.OPT_SERIALIZE_NUMPY
    if caller == "tornado.escape":
        option |= orjson.OPT_NAIVE_UTC | orjson.OPT_UTC_Z
    else:
        option |= orjson.OPT_PASSTHROUGH_DATACLASS
    if sort_keys:
        option |= orjson.OPT_SORT_KEYS
    if indent is not None:
        option |= orjson.OPT_INDENT_2
    if cls is not None:
        _ = cls(
            skipkeys=skipkeys,
            ensure_ascii=ensure_ascii,
            check_circular=check_circular,
            allow_nan=allow_nan,
            indent=indent,
            separators=separators,
            default=default,
            sort_keys=sort_keys,
            **kwargs,
        )
        default = _.default
    output = orjson.dumps(obj, default, option)
    if indent not in {None, 2, "  "}:
        if isinstance(indent, int):
            indent = " " * indent
        indent_bytes = str(indent).encode("UTF-8")
        output = regex.sub(
            rb"(?m)^\s+",
            lambda match: len(match[0]) // 2 * indent_bytes,
            output,
        )
    return output.decode("UTF-8")


def dump(obj: Any, fp: IO[str], **kwargs: Any) -> None:  # noqa: D103
    # pylint: disable=missing-function-docstring
    fp.write(dumps(obj, **kwargs))


def loads(s: str | bytes, **kwargs: Any) -> Any:  # noqa: D103
    # pylint: disable=missing-function-docstring, unused-argument
    return orjson.loads(s)


def load(fp: SupportsRead[str | bytes], **kwargs: Any) -> Any:  # noqa: D103
    # pylint: disable=missing-function-docstring, unused-argument
    return loads(fp.read())



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The tests for the utils module."""

from __future__ import annotations

from urllib.parse import urlsplit

import pytest

from an_website.utils import utils


def test_adding_stuff_to_url() -> None:
    """Test the utils.add_args_to_url function."""
    for url in (
        "https://example.com/",
        "https://example.com/?a=b&b=c",
        "https://example.com/?x=z",
    ):
        assert url == utils.add_args_to_url(url)
        assert "x=y" in urlsplit(utils.add_args_to_url(url, x="y")).query

    assert (
        urlsplit(
            utils.add_args_to_url("https://example.com/", a="b", c="d", e="f")
        ).query
        == "a=b&c=d&e=f"
    )

    assert (
        utils.add_args_to_url("https://example.com/", a="b", c="d", e=True)
        == f"https://example.com/?a=b&c=d&e={utils.bool_to_str(True)}"
    )

    assert (
        f"a={utils.bool_to_str(True)}"
        == urlsplit(utils.add_args_to_url("https://example.com/", a=True)).query
    )


def test_anonomyze_ip() -> None:
    """Test the anonomyze_ip function."""
    assert utils.anonymize_ip("10.88.135.144") == "10.88.135.0"
    assert (
        utils.anonymize_ip("2001:db8:85a3::8a2e:370:7334") == "2001:db8:85a3::"
    )
    with pytest.raises(ValueError):
        utils.anonymize_ip("invalid")
    assert utils.anonymize_ip("invalid", ignore_invalid=True) == "invalid"


def test_bool_str_conversion() -> None:
    """Test the conversion from bool to str and from str to bool."""
    for boolean in (False, True):
        assert boolean is utils.str_to_bool(utils.bool_to_str(boolean))
        assert boolean is utils.str_to_bool(str(boolean))

    for boolean_str in ("sure", "nope"):
        boolean_bool = utils.str_to_bool(boolean_str)
        assert isinstance(boolean_bool, bool)
        assert boolean_str == utils.bool_to_str(boolean_bool)

    with pytest.raises(ValueError):
        utils.str_to_bool("Invalid bool value")

    assert utils.str_to_bool("Invalid bool value", True) is True


def test_country_code_to_flag() -> None:
    """Test the utils.country_code_to_flag function."""
    assert utils.country_code_to_flag("aq") == "🇦🇶"
    assert utils.country_code_to_flag("aQ") == "🇦🇶"
    assert utils.country_code_to_flag("Aq") == "🇦🇶"
    assert utils.country_code_to_flag("AQ") == "🇦🇶"


def test_emojify() -> None:
    """Test the emojify function."""
    assert tuple(utils.emojify("aBc 123 #!*")) == (
        "🇦",
        "🇧",
        "🇨",
        " ",
        "1️⃣",
        "2️⃣",
        "3️⃣",
        " ",
        "#️⃣",
        "❗",
        "*️⃣",
    )
    assert tuple(utils.emojify("!?!?!!")) == ("⁉", "⁉", "‼")
    assert tuple(utils.emojify("Üẞ?!")) == ("🇺", "🇪", "🇸", "🇸", "❓", "❗")
    assert tuple(utils.emojify("2 + 2 - 3 = 0!  ")) == (
        "2️⃣",
        " ",
        "➕",
        " ",
        "2️⃣",
        " ",
        "➖",
        " ",
        "3️⃣",
        " = ",
        "0️⃣",
        "❗",
        "  ",
    )


def test_n_from_set() -> None:
    """Test the n_from_set function."""
    set_ = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12}
    assert len(utils.n_from_set(set_, 3)) == 3
    assert -1 < len(utils.n_from_set(set_, 0)) < 1
    assert -1 < len(utils.n_from_set(set_, -1)) < 1
    for element in utils.n_from_set(set_, 8):
        assert element in set_
    for i in range(len(set_) + 3):
        assert len(utils.n_from_set(set_, i)) == min(len(set_), i)


def test_name_to_id() -> None:
    """Test the name_to_id function."""
    assert utils.name_to_id("     test 42069     ") == "test-42069"
    assert utils.name_to_id("     test-test-42069") == "test-test-42069"
    assert utils.name_to_id("     test_TEST_42069") == "test-test-42069"
    assert utils.name_to_id("     test test 42069") == "test-test-42069"
    assert utils.name_to_id("test__  __TEST_42069") == "test-test-42069"
    assert utils.name_to_id("test test test 42069") == "test-test-test-42069"
    assert utils.name_to_id("TEST test_test_42069") == "test-test-test-42069"
    assert utils.name_to_id("ẞ Ä Ö Ü-ẞÄÖÜ        ") == "ss-ae-oe-ue-ssaeoeue"


def test_replace_umlauts() -> None:
    """Test the replace_umlauts function."""
    assert utils.replace_umlauts("äöüß") == "aeoeuess"
    assert utils.replace_umlauts("ÄÖÜẞ") == "AEOEUESS"
    assert utils.replace_umlauts("Äöüß") == "Aeoeuess"
    assert utils.replace_umlauts("Öäüß") == "Oeaeuess"
    assert utils.replace_umlauts("Üöäß") == "Ueoeaess"
    assert utils.replace_umlauts("ẞäöü") == "SSaeoeue"
    assert utils.replace_umlauts("ẞäöü ÄÖÜẞ") == "SSaeoeue AEOEUESS"
    assert utils.replace_umlauts("ẞäöü ÄÖÜß") == "SSaeoeue AeOeUess"
    assert utils.replace_umlauts("ẞÄÖÜ ÄöÜß") == "SSAEOEUE AeoeUess"


def test_time_to_str() -> None:
    """Test the time_to_str function."""
    assert utils.time_to_str(0) == "0d 0h 0min 0s"
    assert utils.time_to_str(133769420) == "1548d 6h 10min 20s"


def test_get_close_matches() -> None:
    """Test the get_close_matches function."""
    assert utils.get_close_matches("a", ("a", "b"), cutoff=0) == ("a",)
    assert utils.get_close_matches("", ("a", "b"), cutoff=1.0) == ("a", "b")
    assert utils.get_close_matches("aa", ("ab", "baa", "acab")) == (
        "baa",
        "ab",
        "acab",
    )
    assert utils.get_close_matches("ab", "bcdefghiajklmnop") == ("a", "b")
    assert utils.get_close_matches("a", "awawawawawawawawawawawa", 1) == ("a",)
    assert utils.get_close_matches("a𓆗", "a") == ("a",)


if __name__ == "__main__":
    test_adding_stuff_to_url()
    test_anonomyze_ip()
    test_bool_str_conversion()
    test_country_code_to_flag()
    test_emojify()
    test_n_from_set()
    test_name_to_id()
    test_replace_umlauts()
    test_time_to_str()
    test_get_close_matches()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A page that tells users their IP address."""

from __future__ import annotations

from ipaddress import ip_address

from ..utils.request_handler import APIRequestHandler, HTMLRequestHandler
from ..utils.utils import ModuleInfo


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=(
            (r"/ip", IP),
            (r"/api/ip", IPAPI),
        ),
        name="IP-Informationen",
        description="Erfahre deine IP-Adresse",
        path="/ip",
        keywords=(
            "IP-Adresse",
            "Was ist meine IP?",
        ),
    )


class IPAPI(APIRequestHandler):
    """The request handler for the IP API."""

    async def get(self, *, head: bool = False) -> None:
        """Handle GET requests to the IP API."""
        if head:
            return

        if geoip := await self.geoip():
            return await self.finish_dict(
                ip=self.request.remote_ip,
                country=geoip.get("country_flag"),
            )

        return await self.finish_dict(ip=self.request.remote_ip)


class IP(HTMLRequestHandler):
    """The request handler for the IP page."""

    async def get(self, *, head: bool = False) -> None:
        """Handle GET requests to the IP page."""
        if head:
            return

        if not self.request.remote_ip:
            return await self.render(
                "pages/empty.html",
                text="IP-Adresse konnte nicht ermittelt werden.",
            )

        await self.render(
            "pages/ip.html",
            ip=self.request.remote_ip,
            flag=(
                "🔁"
                if ip_address(self.request.remote_ip).is_loopback
                else (await self.geoip() or {}).get("country_flag") or "❔"
            ),
        )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The tests for the contact page."""

from __future__ import annotations

import asyncio
import random
import sys
from email.message import Message

import orjson as json
import pytest
from tornado.httpclient import AsyncHTTPClient

from an_website.contact import contact


def test_add_geoip_info_to_message() -> None:
    """Test the add_geoip_info_to_message function."""
    geoip = {
        "continent_name": "Oceania",
        "country_name": "Australia",
        "location": {"lon": 143.2104, "lat": -33.494},
        "country_iso_code": "AU",
        "timezone": "Australia/Sydney",
    }

    message = Message()
    contact.add_geoip_info_to_message(message, geoip)

    # pylint: disable=protected-access
    assert message._headers == [  # type: ignore[attr-defined]
        ("X-GeoIP-continent-name", "Oceania"),
        ("X-GeoIP-country-name", "Australia"),
        ("X-GeoIP-location-lon", "143.2104"),
        ("X-GeoIP-location-lat", "-33.494"),
        ("X-GeoIP-country-iso-code", "AU"),
        ("X-GeoIP-timezone", "Australia/Sydney"),
    ]


def test_add_geoip_info_to_message_recursive() -> None:
    """Test infinite recursion."""
    # pylint: disable=dict-init-mutate
    geoip = {"spam": "eggs"}
    geoip["sausage"] = geoip  # type: ignore[assignment]
    message = Message()
    with pytest.raises(RecursionError):
        contact.add_geoip_info_to_message(message, geoip)


async def test_sending_email() -> None:
    """Test sending emails."""
    user = f"{random.randbytes(10).hex()}"
    subject = random.randbytes(10).hex()
    content = random.randbytes(20).hex() + "\n"
    message = Message()
    message["Subject"] = subject
    message.set_payload(content)
    await asyncio.to_thread(
        contact.send_message,
        message=message,
        from_address="Marcell D'Avis <davis@1und1.de>",
        recipients=(f"an-website <{user}@restmail.net>",),
        server="restmail.net",
        port=25,
    )

    response = await AsyncHTTPClient().fetch(
        f"https://restmail.net/mail/{user}"
    )

    for email in json.loads(response.body):
        if (  # pylint: disable=too-many-boolean-expressions
            email["subject"] == subject
            and email["text"] == content
            and email["from"][0]["name"] == "Marcell D'Avis"
            and email["from"][0]["address"] == "davis@1und1.de"
            and email["to"][0]["name"] == "an-website"
            and email["to"][0]["address"] == f"{user}@restmail.net"
        ):
            break
    else:
        import __future__  # pylint: disable=import-outside-toplevel

        assert __future__.barry_as_FLUFL.mandatory < sys.version_info  # type: ignore[attr-defined]  # noqa: B950


if __name__ == "__main__":
    test_add_geoip_info_to_message()
    test_add_geoip_info_to_message_recursive()
    asyncio.run(test_sending_email())



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A page to create new wrong quotes."""

from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Final, cast

from tornado.web import HTTPError, MissingArgumentError

from ..utils.data_parsing import parse_args
from ..utils.utils import bounded_edit_distance
from .utils import (
    AUTHORS_CACHE,
    QUOTES_CACHE,
    Author,
    Quote,
    QuoteReadyCheckHandler,
    fix_author_name,
    fix_quote_str,
    make_api_request,
    parse_author,
    parse_quote,
)

LOGGER: Final = logging.getLogger(__name__)


async def create_quote(quote_str: str, author: Author) -> Quote:
    """Create a quote."""
    quote_str = fix_quote_str(quote_str)

    quote = get_quote_by_str(quote_str)
    if quote is not None:
        return quote

    result = parse_quote(
        await make_api_request(
            "quotes",
            method="POST",
            body={
                "author": str(author.id),
                "quote": quote_str,
            },
            entity_should_exist=False,
        )
    )

    LOGGER.info("Created quote %d: %r", result.id, result.quote)

    return result


async def create_author(author_str: str) -> Author:
    """Create an author."""
    author_str = fix_author_name(author_str)

    author = get_author_by_name(author_str)
    if author is not None:
        return author

    result = parse_author(
        await make_api_request(
            "authors",
            method="POST",
            body={"author": author_str},
            entity_should_exist=False,
        )
    )

    LOGGER.info("Created author %d: %r", result.id, result.name)

    return result


async def create_wrong_quote(
    real_author_param: Author | str,
    fake_author_param: Author | str,
    quote_param: Quote | str,
) -> str:
    """Create a wrong quote and return the id in the q_id-a_id format."""
    if isinstance(fake_author_param, str):
        if not fake_author_param:
            raise HTTPError(400, "Fake author is needed, but empty.")
        fake_author = await create_author(fake_author_param)
    else:
        fake_author = fake_author_param

    if isinstance(quote_param, str):
        if not quote_param:
            raise HTTPError(400, "Quote is needed, but empty.")

        if isinstance(real_author_param, str):
            if not real_author_param:
                raise HTTPError(400, "Real author is needed, but empty.")
            real_author = await create_author(real_author_param)
        else:
            real_author = real_author_param
        quote = await create_quote(quote_param, real_author)
    else:
        quote = quote_param

    return f"{quote.id}-{fake_author.id}"


async def get_authors(author_name: str) -> list[Author | str]:
    """Get the possible meant authors based on the str."""
    author = get_author_by_name(author_name)
    if author is not None:
        return [author]

    author_name_lower = author_name.lower()
    max_distance = min(5, len(author_name) // 2 + 1)
    authors: list[Author | str] = [
        *(
            author
            for author in AUTHORS_CACHE.values()
            if bounded_edit_distance(
                author.name.lower(), author_name_lower, max_distance + 1
            )
            <= max_distance
        ),
        fix_author_name(author_name),
    ]
    # authors should be in most cases title case
    fixed_author = author_name.title()
    if fixed_author not in authors:
        authors.append(fixed_author)
    # no other fixes for authors that are less long
    if len(author_name) < 2:
        return authors
    # maybe only the first letter upper case
    fixed_author_2 = author_name[0].upper() + author_name[1:]
    if fixed_author_2 not in authors:
        authors.append(fixed_author_2)
    return authors


def get_author_by_name(name: str) -> None | Author:
    """Get an author by its name."""
    lower_name = fix_author_name(name).lower()
    for author in AUTHORS_CACHE.values():
        if author.name.lower() == lower_name:
            return author
    return None


def get_quote_by_str(quote_str: str) -> None | Quote:
    """Get an author by its name."""
    lower_quote = fix_quote_str(quote_str).lower()
    for quote in QUOTES_CACHE.values():
        if quote.quote.lower() == lower_quote:
            return quote
    return None


async def get_quotes(quote_str: str) -> list[Quote | str]:
    """Get the possible meant quotes based on the str."""
    quote: None | Quote = get_quote_by_str(quote_str)
    if isinstance(quote, Quote):
        return [quote]

    lower_quote_str = quote_str.lower()
    max_distance = min(16, len(quote_str) // 2 + 1)
    quotes: list[Quote | str] = [
        *(
            quote
            for quote in QUOTES_CACHE.values()
            if bounded_edit_distance(
                quote.quote.lower(), lower_quote_str, max_distance + 1
            )
            <= max_distance
        ),
        fix_quote_str(quote_str),
    ]
    if not (
        quote_str.endswith("?")
        or quote_str.endswith(".")
        or quote_str.endswith("!")
    ):
        quotes.append(quote_str + ".")
    return quotes


@dataclass(slots=True, frozen=True)
class QuoteInfoArgs:
    """Class representing a quote id and an author id."""

    quote: int | None = None
    author: int | None = None


class CreatePage1(QuoteReadyCheckHandler):
    """The request handler for the create page."""

    RATELIMIT_POST_LIMIT = 5
    RATELIMIT_POST_COUNT_PER_PERIOD = 10

    @parse_args(type_=QuoteInfoArgs)
    async def get(self, *, args: QuoteInfoArgs, head: bool = False) -> None:
        """Handle GET requests to the create page."""
        if args.quote is not None and args.author is not None:
            return self.redirect(f"/zitate/{args.quote}-{args.author}")

        if head:
            return

        await self.render(
            "pages/quotes/create1.html",
            quotes=tuple(QUOTES_CACHE.values()),
            authors=tuple(AUTHORS_CACHE.values()),
            selected_quote=(
                None if args.quote is None else QUOTES_CACHE.get(args.quote)
            ),
            selected_author=(
                None if args.author is None else AUTHORS_CACHE.get(args.author)
            ),
        )

    async def post(self) -> None:
        """Handle POST requests to the create page."""
        if self.get_argument("keine-aktion", ""):
            LOGGER.info(
                "Versuch Zitat zu erstellen mit keine-aktion Parameter.\n"
                "Pfad: %s\nBody: %s",
                self.request.path,
                self.request.body,
            )
            await self.render(
                "pages/empty.html",
                text="Kein Zitat erstellt.",
            )
            return

        quote_str = self.get_argument("quote-1", None)
        fake_author_str = self.get_argument("fake-author-1", None)
        if not (quote_str and fake_author_str):
            raise HTTPError(
                400,
                reason=(
                    "Missing arguments. quote-1 and fake-author-1 are needed."
                ),
            )
        quote: None | Quote = get_quote_by_str(quote_str)
        fake_author: None | Author = get_author_by_name(fake_author_str)

        if quote and fake_author:
            # if selected existing quote and existing
            # fake author just redirect to the page of this quote
            return self.redirect(
                self.fix_url(f"/zitate/{quote.id}-{fake_author.id}"),
                status=303,
            )

        if not quote:
            # TODO: search for real author, to reduce work for users
            real_author_str = self.get_argument("real-author-1", None)
            if not real_author_str:
                raise HTTPError(
                    400, reason="Missing arguments. real-author-1 is needed."
                )

        quotes: list[Quote | str] = (
            [quote] if quote else await get_quotes(quote_str)
        )
        real_authors: list[Author | str] = (
            [quote.author]
            if quote
            else await get_authors(cast(str, real_author_str))  # type: ignore[possibly-undefined]  # noqa: B950
        )
        fake_authors: list[Author | str] = (
            [fake_author] if fake_author else await get_authors(fake_author_str)
        )

        if 1 == len(quotes) == len(real_authors) == len(fake_authors):
            LOGGER.info(
                "Creating wrong quote using existing: %r (%r) - %r",
                quotes[0],
                real_authors[0],
                fake_authors[0],
            )

            wq_id = await create_wrong_quote(
                real_authors[0],
                fake_authors[0],
                quotes[0],
            )
            return self.redirect(
                self.fix_url(f"/zitate/{wq_id}"),
                status=303,
            )

        await self.render(
            "pages/quotes/create2.html",
            quotes=quotes,
            real_authors=real_authors,
            fake_authors=fake_authors,
        )


class CreatePage2(QuoteReadyCheckHandler):
    """The request handler for the second part of the create page."""

    RATELIMIT_POST_LIMIT = 2
    RATELIMIT_POST_COUNT_PER_PERIOD = 1
    RATELIMIT_POST_PERIOD = 45

    async def post(self) -> None:
        """Handle POST requests to the create page."""
        if self.get_argument("nicht-erstellen", ""):
            LOGGER.info(
                "Versuch Zitat zu erstellen mit nicht-erstellen Parameter.\n"
                "Pfad: %s\nBody: %s",
                self.request.path,
                self.request.body,
            )
            await self.render(
                "pages/empty.html",
                text="Kein Zitat erstellt.",
            )
            return

        quote_str = self.get_argument("quote-2", None)
        if not quote_str:
            raise MissingArgumentError("quote-2")
        fake_author_str = self.get_argument("fake-author-2", None)
        if not fake_author_str:
            raise MissingArgumentError("fake-author-2")

        if (quote := get_quote_by_str(quote_str)) and (
            fake_author := get_author_by_name(fake_author_str)
        ):
            # if selected existing quote and existing
            # fake author just redirect to the page of this quote
            return self.redirect(
                self.fix_url(f"/zitate/{quote.id}-{fake_author.id}"),
                status=303,
            )

        real_author = self.get_argument("real-author-2", None)
        if not real_author:
            raise MissingArgumentError("real-author-2")

        LOGGER.info(
            "Creating wrong quote: %r (%r) - %r",
            quote_str,
            real_author,
            fake_author_str,
        )

        wq_id = await create_wrong_quote(
            real_author,
            fake_author_str,
            quote_str,
        )
        return self.redirect(
            self.fix_url(
                f"/zitate/{wq_id}",
            ),
            status=303,
        )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""A provider of the Redis PubSub class."""

from __future__ import annotations

import logging
from collections.abc import Collection
from dataclasses import dataclass
from typing import Any, Final

from redis.asyncio.client import PubSub, Redis

from .. import EVENT_REDIS

LOGGER: Final = logging.getLogger(__name__)


@dataclass(slots=True)
class PubSubProvider:
    """Provide a PubSub object."""

    channels: Collection[str]
    settings: dict[str, Any]
    worker: int | None
    _ps: PubSub | None = None
    _redis: Redis[str] | None = None

    async def __call__(self) -> PubSub:
        """Get PubSub object."""
        if not self.settings.get("REDIS"):
            LOGGER.error("Redis not available on worker %s", self.worker)

        await EVENT_REDIS.wait()

        redis: Redis[str] = self.settings["REDIS"]

        if self._ps:
            if self._redis == redis:
                return self._ps
            LOGGER.info(
                "Closing old PubSub connection on worker %s", self.worker
            )
            await self._ps.close()

        self._ps = redis.pubsub()
        self._redis = redis
        await self._ps.subscribe(*self.channels)
        return self._ps



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Create the required stuff for the soundboard from the info in info.json."""

from __future__ import annotations

from typing import Final

from tornado.web import RedirectHandler

from .. import DIR as ROOT_DIR
from ..utils.static_file_from_traversable import TraversableStaticFileHandler
from ..utils.utils import ModuleInfo, PageInfo
from .soundboard import SoundboardHTMLHandler, SoundboardRSSHandler

DIR: Final = ROOT_DIR / "soundboard"


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        name="Känguru-Soundboard",
        short_name="Soundboard",
        description=(
            "Ein Soundboard mit coolen Sprüchen und Sounds aus den "
            "Känguru-Chroniken"
        ),
        path="/soundboard",
        keywords=("Soundboard", "Känguru", "Witzig", "Sprüche"),
        handlers=(
            (
                r"/soundboard/files/(.*mp3)",
                TraversableStaticFileHandler,
                {"root": DIR / "files", "hashes": {}},
            ),
            (
                r"/soundboard/feed",
                SoundboardRSSHandler,
            ),
            (
                r"/soundboard/feed\.(rss|xml)",
                RedirectHandler,
                {"url": "/soundboard/feed"},
            ),
            (  # redirect handler for legacy reasons
                r"/soundboard/k(ä|%C3%A4)nguru(/.+|)",
                RedirectHandler,
                {"url": "/soundboard/kaenguru{1}"},
            ),
            (
                r"/soundboard/([^./]+)/feed",
                SoundboardRSSHandler,
            ),
            (
                r"/soundboard/([^/]+)(\.(rss|xml)|/feed\.(rss|xml))",
                RedirectHandler,
                {"url": "/soundboard/{0}/feed"},
            ),
            (
                r"/soundboard",
                SoundboardHTMLHandler,
            ),
            (
                r"/soundboard/([^./]+)",
                SoundboardHTMLHandler,
            ),
        ),
        sub_pages=(
            PageInfo(
                name="Soundboard-Suche",
                description="Durchsuche das Soundboard",
                path="/soundboard/suche",
                keywords=("Suche",),
            ),
            PageInfo(
                name="Soundboard-Personen",
                description="Das Soundboard mit Sortierung nach Personen",
                path="/soundboard/personen",
                keywords=("Personen",),
            ),
        ),
        aliases=(
            "/kaenguru-soundboard",
            "/känguru-soundboard",
            "/k%C3%A4nguru-soundboard",
        ),
    )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Config file for the page that swaps words."""

from __future__ import annotations

from dataclasses import dataclass, field
from functools import lru_cache
from typing import TypeAlias

import regex
from regex import Match, Pattern


def copy_case_letter(char_to_steal_case_from: str, char_to_change: str) -> str:
    """
    Copy the case of one string to another.

    This method assumes that the whole string has the same case,
    like it is the case for a letter.
    """
    return (
        char_to_change.upper()  # char_to_steal_case_from is upper case
        if char_to_steal_case_from.isupper()
        else char_to_change.lower()  # char_to_steal_case_from is lower case
    )


def copy_case(reference_word: str, word_to_change: str) -> str:
    """Copy the case of one string to another."""
    # lower case: "every letter is lower case"
    if reference_word.islower():
        return word_to_change.lower()
    # upper case: "EVERY LETTER IS UPPER CASE"
    if reference_word.isupper():
        return word_to_change.upper()
    # title case: "Every Word Begins With Upper Case Letter"
    if reference_word.istitle():
        return word_to_change.title()

    split_ref = reference_word.split(" ")
    split_word = word_to_change.split(" ")
    # if both equal length and not len == 1
    if len(split_ref) == len(split_word) != 1:
        # go over every word, if there are spaces
        return " ".join(
            copy_case(split_ref[i], split_word[i])
            for i in range(len(split_ref))
        )

    # other words
    new_word: list[str] = []  # use list for speed
    for i, letter in enumerate(word_to_change):
        new_word.append(
            copy_case_letter(
                # overflow original word for mixed case
                reference_word[i % len(reference_word)],
                letter,
            )
        )
    # create new word and return it
    return "".join(new_word)


class ConfigLine:  # pylint: disable=too-few-public-methods
    """Class used to represent a word pair."""

    def to_conf_line(self, len_of_left: None | int = None) -> str:
        """Get how this would look like in a config."""
        raise NotImplementedError


@dataclass(frozen=True, slots=True)
class Comment(ConfigLine):
    """Class used to represent a comment."""

    comment: str

    def to_conf_line(self, len_of_left: None | int = None) -> str:
        """Get how this would look like in a config."""
        return "" if not self.comment else f"# {self.comment}"


@dataclass(frozen=True, init=False, slots=True)
class WordPair(ConfigLine):
    """Parent class representing a word pair."""

    word1: str
    word2: str
    # separator between the two words, that shouldn't be changed
    separator: str = field(default="", init=False)

    def get_replacement(self, word: str) -> str:
        """Get the replacement for a given word with the same case."""
        raise NotImplementedError

    def len_of_left(self) -> int:
        """Get the length to the left of the separator."""
        return len(self.word1)

    def to_conf_line(self, len_of_left: None | int = None) -> str:
        """Get how this would look like in a config."""
        left, separator, right = self.word1, self.separator, self.word2
        if len_of_left is None:
            return "".join((left, separator.strip(), right))
        return (
            left
            + (" " * (1 + len_of_left - self.len_of_left()))
            + separator
            + " "
            + right
        )

    def to_pattern_str(self) -> str:
        """Get the pattern that matches the replaceable words."""
        raise NotImplementedError


@dataclass(frozen=True, slots=True)
class OneWayPair(WordPair):
    """Class used to represent a word that gets replaced with another."""

    # separator between the two words, that shouldn't be changed
    separator: str = field(default=" =>", init=False)

    def get_replacement(self, word: str) -> str:
        """Get the replacement for a given word with the same case."""
        if regex.fullmatch(self.word1, word) is not None:
            return self.word2
        if regex.fullmatch(self.word1, word, regex.IGNORECASE) is not None:
            return copy_case(word, self.word2)
        return word

    def to_pattern_str(self) -> str:
        """Get the pattern that matches the replaceable words."""
        return self.word1


@dataclass(frozen=True, slots=True)
class TwoWayPair(WordPair):
    """Class used to represent two words that get replaced with each other."""

    # separator between the two words, that shouldn't be changed
    separator: str = field(default="<=>", init=False)

    def get_replacement(self, word: str) -> str:
        """Get the replacement for a given word with the same case."""
        if self.word1 == word:
            return self.word2
        if self.word2 == word:
            return self.word1
        # doesn't match case sensitive
        word_lower = word.lower()
        if self.word1.lower() == word_lower:
            return copy_case(word, self.word2)
        if self.word2.lower() == word_lower:
            return copy_case(word, self.word1)
        return word

    def to_pattern_str(self) -> str:
        """Get the pattern that matches the replaceable words."""
        return f"{self.word1}|{self.word2}"


WORDS_TUPLE: TypeAlias = tuple[ConfigLine, ...]  # pylint: disable=invalid-name

LINE_END_REGEX: Pattern[str] = regex.compile(
    r"(?m)[\n;]",  # ";" or new line
)

COMMENT_LINE_REGEX: Pattern[str] = regex.compile(
    r"#"  # the start of the comment
    r"\s*"  # optional white space
    r"("  # start of the group
    r"[^\s]"  # a non white space char (start of comment)
    r".*"  # whatever
    r")?"  # end group, make it optional, to allow lines with only #
)

LINE_REGEX: Pattern[str] = regex.compile(
    r"\s*"  # white spaces to strip the word
    r"("  # start group one for the first word
    r"[^\s<=>]"  # the start of the word; can't contain: \s, "<", "=", ">"
    r"[^<=>]*"  # the middle of the word; can't contain: "<", "=", ">"="
    r"[^\s<=>]"  # the end of the word; can't contain: \s, "<", "=", ">"
    r"|[^\s<=>]?"  # one single letter word; optional -> better error message
    r")"  # end group one for the first word
    r"\s*"  # white spaces to strip the word
    r"(<?=>)"  # the seperator in the middle either "=>" or "<=>"
    r"\s*"  # white spaces to strip the word
    r"("  # start group two for the second word
    r"[^\s<=>]"  # the start of the word; can't contain: \s, "<", "=", ">"
    r"[^<=>]*"  # the middle of the word; can't contain: "<", "=", ">"
    r"[^\s<=>]"  # the end of the word; can't contain: \s, "<", "=", ">"
    r"|[^\s<=>]?"  # one single letter word; optional -> better error message
    r")"  # end group two for the second word
    r"\s*"  # white spaces to strip the word
)


@lru_cache(20)
def parse_config_line(  # pylint: disable=too-complex
    line: str, line_num: int = -1
) -> ConfigLine:
    """Parse one config line to one ConfigLine instance."""
    # remove white spaces to fix stuff, behaves weird otherwise
    line = line.strip()

    if not line:
        return Comment("")  # empty comment → empty line

    if match := regex.fullmatch(COMMENT_LINE_REGEX, line):
        return Comment(match[1])

    match = regex.fullmatch(LINE_REGEX, line)
    if match is None:
        raise InvalidConfigError(line_num, line, "Line is invalid.")

    left, separator, right = match[1], match[2], match[3]
    if not left:
        raise InvalidConfigError(line_num, line, "Left of separator is empty.")
    if separator not in ("<=>", "=>"):
        raise InvalidConfigError(
            line_num, line, "No separator ('<=>' or '=>') present."
        )
    if not right:
        raise InvalidConfigError(line_num, line, "Right of separator is empty.")

    try:
        # compile to make sure it doesn't break later
        left_re = regex.compile(left)
    except regex.error as exc:
        raise InvalidConfigError(
            line_num, line, f"Left is invalid regex: {exc}"
        ) from exc

    if separator == "<=>":
        try:
            # compile to make sure it doesn't break later
            right_re = regex.compile(right)
        except regex.error as exc:
            raise InvalidConfigError(
                line_num, line, f"Right is invalid regex: {exc}"
            ) from exc
        # pylint: disable=too-many-function-args
        return TwoWayPair(left_re.pattern, right_re.pattern)
    if separator == "=>":
        # pylint: disable=too-many-function-args
        return OneWayPair(left_re.pattern, right)

    raise InvalidConfigError(line_num, line, "Something went wrong.")


@dataclass(frozen=True, slots=True)
class InvalidConfigError(Exception):
    """Exception raised if the config is invalid."""

    line_num: int
    line: str
    reason: str

    def __str__(self) -> str:
        """Exception to str."""
        return (
            f"Error in line {self.line_num}: {self.line.strip()!r} "
            f"with reason: {self.reason}"
        )


class SwappedWordsConfig:  # pylint: disable=eq-without-hash
    """SwappedWordsConfig class used to swap words in strings."""

    def __eq__(self, other: object) -> bool:
        """Check equality based on the lines."""
        if not isinstance(other, SwappedWordsConfig):
            return NotImplemented
        return self.lines == other.lines

    def __init__(self, config: str):
        """Parse a config string to a instance of this class."""
        self.lines: WORDS_TUPLE = tuple(
            parse_config_line(line, i)
            for i, line in enumerate(
                regex.split(LINE_END_REGEX, config.strip())
            )
            if line
        )

    def get_regex(self) -> Pattern[str]:
        """Get the regex that matches every word in this."""
        return regex.compile(
            "|".join(
                tuple(
                    f"(?P<n{i}>{word_pair.to_pattern_str()})"
                    for i, word_pair in enumerate(self.lines)
                    if isinstance(word_pair, WordPair)
                )
            ),
            regex.IGNORECASE,
        )

    def get_replaced_word(self, match: Match[str]) -> str:
        """Get the replaced word with the same case as the match."""
        for key, word in match.groupdict().items():
            if isinstance(word, str) and key.startswith("n"):
                return self.get_replacement_by_group_name(key, word)
        # if an unknown error happens return the match to change nothing
        return match[0]

    def get_replacement_by_group_name(self, group_name: str, word: str) -> str:
        """Get the replacement of a word by the group name it matched."""
        if not group_name.startswith("n") or group_name == "n":
            return word
        index = int(group_name[1:])
        config_line = self.lines[index]
        if isinstance(config_line, WordPair):
            return config_line.get_replacement(word)

        return word

    def swap_words(self, text: str) -> str:
        """Swap the words in the text."""
        return self.get_regex().sub(self.get_replaced_word, text)

    def to_config_str(self, minified: bool = False) -> str:
        """Create a readable config str from this."""
        if not self.lines:
            return ""
        if minified:
            return ";".join(
                word_pair.to_conf_line()
                for word_pair in self.lines
                if isinstance(word_pair, WordPair)
            )
        max_len = max(
            (
                word_pair.len_of_left()
                for word_pair in self.lines
                if isinstance(word_pair, WordPair)
            ),
            default=0,
        )
        return "\n".join(
            word_pair.to_conf_line(max_len) for word_pair in self.lines
        )


def minify(config: str) -> str:
    """Minify a config string."""
    return SwappedWordsConfig(config).to_config_str(True)


def beautify(config: str) -> str:
    """Beautify a config string."""
    return SwappedWordsConfig(config).to_config_str()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""EXAMPLE."""

from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Final

from tornado.web import MissingArgumentError

from ..utils.data_parsing import parse_args
from ..utils.request_handler import APIRequestHandler, HTMLRequestHandler
from ..utils.utils import ModuleInfo

LOGGER: Final = logging.getLogger(__name__)


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=(
            (r"/beispiel", Example),
            (r"/api/beispiel", ExampleAPI),
        ),
        name="EXAMPLE",
        short_name="EXAMPLE",
        description="EXAMPLE",
        path="/beispiel",
        aliases=("/example",),
        sub_pages=(),
        keywords=(),
    )


@dataclass(slots=True)
class ExampleArguments:
    """The arguments for the example page."""

    name: str = "Welt"

    def validate(self) -> None:
        """Validate this."""
        self.name = self.name.strip()
        if not self.name:
            raise MissingArgumentError("name")


class Example(HTMLRequestHandler):
    """The request handler for the example page."""

    @parse_args(type_=ExampleArguments, validation_method="validate")
    async def get(self, *, args: ExampleArguments, head: bool = False) -> None:
        """Handle GET requests to the page."""
        self.set_header("X-Name", args.name)
        if head:
            # only after all headers have been set and the status code is clear
            return
        await self.render("pages/EXAMPLE.html", name=args.name)


class ExampleAPI(APIRequestHandler):
    """The request handler for the example API."""

    @parse_args(type_=ExampleArguments, validation_method="validate")
    async def get(self, *, args: ExampleArguments, head: bool = False) -> None:
        """Handle GET requests to the API."""
        # pylint: disable=unused-argument
        await self.finish({"text": f"Hallo, {args.name}!", "name": args.name})



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The uptime page that shows the time the website is running."""

from __future__ import annotations

import logging
import math
import time
from typing import Final, TypedDict

import regex
from elasticsearch import AsyncElasticsearch
from tornado.web import HTTPError, RedirectHandler

from .. import EPOCH, EVENT_ELASTICSEARCH, NAME, UPTIME
from ..utils.base_request_handler import BaseRequestHandler
from ..utils.request_handler import APIRequestHandler, HTMLRequestHandler
from ..utils.utils import ModuleInfo, time_to_str

LOGGER: Final = logging.getLogger(__name__)


class AvailabilityDict(TypedDict):  # noqa: D101
    # pylint: disable=missing-class-docstring
    up: int
    down: int
    total: int
    percentage: None | float


def get_module_info() -> ModuleInfo:
    """Create and return the ModuleInfo for this module."""
    return ModuleInfo(
        handlers=(
            (r"/betriebszeit", UptimeHandler),
            (r"/betriebszeit/verfuegbarkeit.svg", AvailabilityChartHandler),
            (r"/api/betriebszeit", UptimeAPIHandler),
            (r"/api/uptime/*", RedirectHandler, {"url": "/api/betriebszeit"}),
        ),
        name="Betriebszeit",
        description="Die Dauer, die die Webseite am Stück in Betrieb ist",
        path="/betriebszeit",
        aliases=("/uptime",),
        keywords=("Uptime", "Betriebszeit", "Zeit"),
    )


class EsAvailabilityKwargs(TypedDict):
    """Subset of kwargs to Elasticsearch search."""

    index: str
    query: dict[str, dict[str, list[dict[str, dict[str, dict[str, str]]]]]]
    size: int
    aggs: dict[str, dict[str, dict[str, str]]]


ES_AVAILABILITY_KWARGS: Final[EsAvailabilityKwargs] = {
    "index": "heartbeat-*,synthetics-*",
    "query": {
        "bool": {
            "filter": [
                {"range": {"@timestamp": {"gte": "now-1M"}}},
                {"term": {"monitor.type": {"value": "http"}}},
                {
                    "term": {
                        "service.name": {"value": NAME.removesuffix("-dev")}
                    }
                },
            ]
        }
    },
    "size": 0,
    "aggs": {
        "up": {"sum": {"field": "summary.up"}},
        "down": {"sum": {"field": "summary.down"}},
    },
}


async def get_availability_data(
    elasticsearch: AsyncElasticsearch,
) -> None | tuple[int, int]:  # (up, down)
    """Get the availability data."""
    try:
        data = await elasticsearch.search(**ES_AVAILABILITY_KWARGS)
    except Exception:  # pylint: disable=broad-exception-caught
        LOGGER.exception("Getting availability data from Elasticsearch failed.")
        return None
    data.setdefault("aggregations", {"up": {"value": 0}, "down": {"value": 0}})
    return (
        int(data["aggregations"]["up"]["value"]),
        int(data["aggregations"]["down"]["value"]),
    )


def get_availability_dict(up: int, down: int) -> AvailabilityDict:
    """Get the availability data as a dict."""
    return {
        "up": up,
        "down": down,
        "total": up + down,
        "percentage": 100 * up / (up + down) if up + down else None,
    }


AVAILABILITY_CHART: Final[str] = regex.sub(
    r"\s+",
    " ",
    """
<svg height="20"
     width="20"
     viewBox="0 0 20 20"
     xmlns="http://www.w3.org/2000/svg"
><circle r="10" cx="10" cy="10" fill="red" />
    <circle r="5" cx="10" cy="10" fill="transparent"
        stroke="green"
        stroke-width="10"
        stroke-dasharray="%2.2f 31.4159"
        transform="rotate(-90) translate(-20)" />
</svg>
""".strip(),
)


class UptimeHandler(HTMLRequestHandler):
    """The request handler for the uptime page."""

    COMPUTE_ETAG = False
    POSSIBLE_CONTENT_TYPES = (
        *HTMLRequestHandler.POSSIBLE_CONTENT_TYPES,
        *APIRequestHandler.POSSIBLE_CONTENT_TYPES,
    )

    async def get(self, *, head: bool = False) -> None:
        """Handle GET requests."""
        self.set_header("Cache-Control", "no-cache")
        if head:
            return
        if self.content_type in APIRequestHandler.POSSIBLE_CONTENT_TYPES:
            return await self.finish(await self.get_uptime_data())
        await self.render("pages/uptime.html", **(await self.get_uptime_data()))

    async def get_uptime_data(
        self,
    ) -> dict[str, str | float | AvailabilityDict]:
        """Get uptime data."""
        availability_data = (
            await get_availability_data(self.elasticsearch)
            if EVENT_ELASTICSEARCH.is_set()
            else None
        ) or (0, 0)
        return {
            "uptime": (uptime := UPTIME.get()),
            "uptime_str": time_to_str(uptime),
            "start_time": time.time() - uptime - EPOCH,
            "availability": get_availability_dict(*availability_data),
        }


class AvailabilityChartHandler(BaseRequestHandler):
    """The request handler for the availability chart."""

    POSSIBLE_CONTENT_TYPES = ("image/svg+xml",)
    COMPUTE_ETAG = False

    async def get(self, *, head: bool = False) -> None:
        """Handle GET requests."""
        if not (availability := self.get_argument("a", None)):
            if not EVENT_ELASTICSEARCH.is_set():
                raise HTTPError(503)
            availability_data = await get_availability_data(self.elasticsearch)
            if not availability_data:
                raise HTTPError(503)
            self.redirect(
                self.fix_url(
                    self.request.full_url(),
                    a=int(
                        (
                            get_availability_dict(*availability_data)[
                                "percentage"
                            ]
                            or 0
                        )
                        * 100
                    )
                    / (100 * 100),
                ),
                permanent=False,
            )
            return
        self.set_header(
            "Cache-Control", f"public, min-fresh={60 * 60 * 24 * 14}, immutable"
        )
        if head:
            return
        await self.finish(
            AVAILABILITY_CHART % (math.pi * 10 * float(availability))
        )


class UptimeAPIHandler(APIRequestHandler, UptimeHandler):
    """The request handler for the uptime API."""

    POSSIBLE_CONTENT_TYPES = APIRequestHandler.POSSIBLE_CONTENT_TYPES



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Useful request handlers used by other modules.

This should only contain request handlers and the get_module_info function.
"""

from __future__ import annotations

import logging
from collections.abc import Mapping
from http.client import responses
from typing import Any, ClassVar, Final, override
from urllib.parse import unquote

import regex
from tornado.web import HTTPError

from .base_request_handler import BaseRequestHandler
from .utils import (
    SUS_PATHS,
    get_close_matches,
    remove_suffix_ignore_case,
    replace_umlauts,
)

LOGGER: Final = logging.getLogger(__name__)


class HTMLRequestHandler(BaseRequestHandler):
    """A request handler that serves HTML."""

    POSSIBLE_CONTENT_TYPES: ClassVar[tuple[str, ...]] = (
        "text/html",
        "text/plain",
        "text/markdown",
        "application/vnd.asozial.dynload+json",
    )


class APIRequestHandler(BaseRequestHandler):
    """The base API request handler."""

    POSSIBLE_CONTENT_TYPES: ClassVar[tuple[str, ...]] = (
        "application/json",
        "application/yaml",
    )


class NotFoundHandler(BaseRequestHandler):
    """Show a 404 page if no other RequestHandler is used."""

    @override
    def initialize(self, *args: Any, **kwargs: Any) -> None:
        """Do nothing to have default title and description."""
        if "module_info" not in kwargs:
            kwargs["module_info"] = None
        super().initialize(*args, **kwargs)

    @override
    async def prepare(self) -> None:
        """Throw a 404 HTTP error or redirect to another page."""
        self.now = await self.get_time()

        if self.request.method not in {"GET", "HEAD"}:
            raise HTTPError(404)

        new_path = regex.sub(r"/+", "/", self.request.path.rstrip("/")).replace(
            "_", "-"
        )

        for ext in (".html", ".htm", ".php"):
            new_path = remove_suffix_ignore_case(new_path, f"/index{ext}")
            new_path = remove_suffix_ignore_case(new_path, ext)

        new_path = replace_umlauts(new_path)

        if new_path.lower() in SUS_PATHS:
            self.set_status(469, reason="Nice Try")
            return self.write_error(469)

        if new_path and new_path != self.request.path:
            return self.redirect(self.fix_url(new_path=new_path), True)

        this_path_normalized = unquote(new_path).strip("/").lower()

        paths: Mapping[str, str] = self.settings.get("NORMED_PATHS") or {}

        if p := paths.get(this_path_normalized):
            return self.redirect(self.fix_url(new_path=p), False)

        if len(this_path_normalized) <= 1 and self.request.path != "/":
            return self.redirect(self.fix_url(new_path="/"))

        prefixes = tuple(
            (p, repl)
            for p, repl in paths.items()
            if this_path_normalized.startswith(f"{p}/")
            if f"/{p}" != repl.lower()
            if p != "api"  # api should not be a prefix
        )

        if len(prefixes) == 1:
            ((prefix, replacement),) = prefixes
            return self.redirect(
                self.fix_url(
                    new_path=f"{replacement.strip('/')}"
                    f"{this_path_normalized.removeprefix(prefix)}"
                ),
                False,
            )
        if prefixes:
            LOGGER.error(
                "Too many prefixes %r for path %s", prefixes, self.request.path
            )

        matches = get_close_matches(this_path_normalized, paths, count=1)
        if matches:
            return self.redirect(
                self.fix_url(new_path=paths[matches[0]]), False
            )

        self.set_status(404)
        self.write_error(404)


class ErrorPage(HTMLRequestHandler):
    """A request handler that shows the error page."""

    _success_status: int = 200
    """The status code that is expected to be returned."""

    @override
    def clear(self) -> None:
        """Reset all headers and content for this response."""
        super().clear()
        self._success_status = 200

    @override
    async def get(self, code: str, *, head: bool = False) -> None:
        """Show the error page."""
        # pylint: disable=unused-argument
        status_code = int(code)
        reason = (
            "Nice Try" if status_code == 469 else responses.get(status_code, "")
        )
        # set the status code if it is allowed
        if status_code not in (204, 304) and not 100 <= status_code < 200:
            self.set_status(status_code, reason)
            self._success_status = status_code
        return await self.render(
            "error.html",
            status=status_code,
            reason=reason,
            description=self.get_error_page_description(status_code),
            is_traceback=False,
        )

    @override
    def get_status(self) -> int:
        """Hack the status code.

        This hacks the status code to be 200 if the status code is expected.
        This avoids sending error logs to APM or Webhooks in case of success.

        This depends on the fact that Tornado internally uses self._status_code
        to set the status code in the response and self.get_status() when
        deciding how to log the request.
        """
        status = super().get_status()
        if status == self._success_status:
            return 200
        return status


class ZeroDivision(BaseRequestHandler):
    """A request handler that raises an error."""

    @override
    async def prepare(self) -> None:
        """Divide by zero and raise an error."""
        self.now = await self.get_time()
        self.handle_accept_header(self.POSSIBLE_CONTENT_TYPES)
        if self.request.method != "OPTIONS":
            420 / 0  # pylint: disable=pointless-statement



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""The module containing the impl of the fix_static_path function."""

import logging
import os
from collections.abc import Callable, Iterable, Mapping
from importlib.resources.abc import Traversable
from pathlib import Path
from types import MappingProxyType
from typing import Final

from blake3 import blake3
from openmoji_dist import VERSION as OPENMOJI_VERSION

if "STATIC_DIR" not in locals():
    from .. import STATIC_DIR

LOGGER: Final = logging.getLogger(__name__)


def recurse_directory(
    root: Traversable,
    # pylint: disable-next=redefined-builtin
    filter: Callable[[Traversable], bool] = lambda _: True,
) -> Iterable[str]:
    """Recursively iterate over entries in a directory."""
    dirs: list[str] = ["."]
    while dirs:  # pylint: disable=while-used
        curr_dir = dirs.pop()
        for path in (root if curr_dir == "." else root / curr_dir).iterdir():
            current: str = (
                path.name
                if curr_dir == "."
                else os.path.join(curr_dir, path.name)
            )
            if path.is_dir():
                dirs.append(current)
            if filter(path):
                yield current


def hash_file(path: Traversable) -> str:
    """Hash a file with BLAKE3."""
    hasher = blake3()
    with path.open("rb") as file:
        for data in file:
            hasher.update(data)
    return hasher.hexdigest(8)


def create_file_hashes_dict(
    filter_path_fun: Callable[[str], bool] | None = None
) -> Mapping[str, str]:
    """Create a dict of file hashes."""
    static = Path("/static")
    file_hashes_dict = {
        f"{(static / path).as_posix()}": hash_file(STATIC_DIR / path)
        for path in recurse_directory(STATIC_DIR, lambda path: path.is_file())
        if not path.endswith((".map", ".gz", ".zst"))
        if filter_path_fun is None or filter_path_fun(path)
    }
    if filter_path_fun is None:
        file_hashes_dict["/favicon.png"] = file_hashes_dict[
            "/static/favicon.png"
        ]
        file_hashes_dict["/favicon.jxl"] = file_hashes_dict[
            "/static/favicon.jxl"
        ]
        file_hashes_dict["/humans.txt"] = file_hashes_dict["/static/humans.txt"]
    return MappingProxyType(file_hashes_dict)


def fix_static_path_impl(path: str, file_hashes_dict: Mapping[str, str]) -> str:
    """Fix the path for static files."""
    if not path.startswith("/"):
        path = f"/static/{path}"
    if "?" in path:
        path = path.split("?")[0]
    if path.startswith("/static/openmoji/"):
        return f"{path}?v={OPENMOJI_VERSION}"
    path = path.lower()
    if path in file_hashes_dict:
        hash_ = file_hashes_dict[path]
        return f"{path}?v={hash_}"
    LOGGER.warning("%s not in FILE_HASHES_DICT", path)
    return path
